/*---------------------------------------------------------------------------*/

/*!  \file  frisca_gpu.c
     \brief  a simple slip-spring rheology simulator on a GPU.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include "frisca_gpu.h"

/*---------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
        /*  simulaton parameters
              iterations: number of iterations per one simulator execution
              p: total number of simulation executions
                 (total number of time steps = iterations * p)
              n: number of beads on a chain
              m: number of chains
              n0: bead number between entanglements
              n_s: slip-spring strength parameter
              zeta_s: friction coefficent for a slip-spring
              dt: time step size
              seed: random seed
              kappa: velocity gradient tensor  */
    int iterations = 100;
    int p = 2000;
    int n = 32;
    int m = 512;
    float n0 = 4.0f;
    float n_s = 0.5f;
    float zeta_s = 0.1f;
    float dt = 0.01f;
    float kappa[3][3] = {{0.0f, 0.05f, 0.0f},
                         {0.0f, 0.0f, 0.0f},
                         {0.0f, 0.0f, 0.0f}};
    unsigned long seed = 12398702;

    simulator_state *state;
    int i,z_max;


        /*  initialization  */

    initialize_device(argc,argv);

    z_max = calculate_estimate_for_z_max(n,n0);
    state  = allocate_simulator_state(m,n,z_max);

    initialize_simulator_state(state,seed,n_s,n0);


    set_simulation_parameters(state,n_s,n0,zeta_s,
                              dt,kappa,iterations);

    copy_simulator_state_host_to_device(state);


    for(i = 0;i < p;i++)
    {
        perform_simulation(state);
        calculate_stress_tensor(state);

        printf("%f %f %f %f %f %f\n",
               state->sigma[0],state->sigma[1],state->sigma[2],
               state->sigma[3],state->sigma[4],state->sigma[5]);
    }


    copy_simulator_state_device_to_host(state);

    store_simulator_state_to_file("frisca_gpu_state.h5",state);

        /*  finalization  */

    free_simulator_state(state);

    finalize_device();

    return 0;
}

/*---------------------------------------------------------------------------*/
