/*---------------------------------------------------------------------------*/

/*!  \file  chain.cu
     \brief  definitions and subroutines for slip-springed chains.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <cutil.h>
#include <cutil_inline.h>
#include "chain.h"

/*---------------------------------------------------------------------------*/
__host__ slip_springed_chain *allocate_slip_springed_chain(int m,int n)
{
    slip_springed_chain *chain;

    chain = (slip_springed_chain *)malloc(sizeof(slip_springed_chain));

    chain->m = m;
    chain->n = n;

    chain->r_host = (float *)malloc(3 * m * n * sizeof(float));
    chain->a_host = (float *)malloc(3 * m * n * sizeof(float));
    chain->phi_host = (int *)malloc(3 * m * n * sizeof(int));

    cutilSafeCall(cudaMalloc((void **)&(chain->r_device),
                             3 * m * n * sizeof(float)));
    cutilSafeCall(cudaMalloc((void **)&(chain->a_device),
                             3 * m * n * sizeof(float)));
    cutilSafeCall(cudaMalloc((void **)&(chain->phi_device),
                             m * n * sizeof(int)));

    return chain;
}

/*---------------------------------------------------------------------------*/
__host__ void free_slip_springed_chain(slip_springed_chain *chain)
{
    cutilSafeCall(cudaFree(chain->r_device));
    cutilSafeCall(cudaFree(chain->a_device));
    cutilSafeCall(cudaFree(chain->phi_device));

    free(chain->r_host);
    free(chain->a_host);
    free(chain->phi_host);
    free(chain);
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_slip_springed_chain(slip_springed_chain *chain,
                                             rng_state *rng,float n0,float n_s)
{
    int i,j;
    float *r,*a;
    int *phi;
    double u0,psi0;

    r = chain->r_host;
    a = chain->a_host;
    phi = chain->phi_host;

    u0 = sqrt((double)n_s / 3.0);
    psi0 = exp(-1.0 / (double)n0);

    for(j = 0;j < chain->m;j++)
    {
        r[0] = 0.0f;
        r[1] = 0.0f;
        r[2] = 0.0f;

        for(i = 1;i < chain->n;i++)
        {
            r[3 * i] = r[3 * (i - 1)] + (float)(sqrt(1.0 / 3.0) * nrand(rng));
            r[3 * i + 1] = r[3 * (i - 1) + 1]
                + (float)(sqrt(1.0 / 3.0) * nrand(rng));
            r[3 * i + 2] = r[3 * (i - 1) + 2]
                + (float)(sqrt(1.0 / 3.0) * nrand(rng));
        }

        for(i = 0;i < chain->n;i++)
        {
            if(urand(rng) < psi0)
            {
                phi[i] = 0;
                a[3 * i] = 0.0f;
                a[3 * i + 1] = 0.0f;
                a[3 * i + 2] = 0.0f;
            }
            else
            {
                phi[i] = 1;
                a[3 * i] = r[3 * i] + (float)(u0 * nrand(rng));
                a[3 * i + 1] = r[3 * i + 1] + (float)(u0 * nrand(rng));
                a[3 * i + 2] = r[3 * i + 2] + (float)(u0 * nrand(rng));
            }
        }

        r += 3 * chain->n;
        a += 3 * chain->n;
        phi += chain->n;
    }
}

/*---------------------------------------------------------------------------*/
__host__ void copy_slip_springed_chain_host_to_device(slip_springed_chain *chain)
{
    cutilSafeCall(cudaMemcpy(chain->r_device, chain->r_host,
                             3 * chain->m * chain->n * sizeof(float),
                             cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(chain->a_device, chain->a_host,
                             3 * chain->m * chain->n * sizeof(float),
                             cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(chain->phi_device, chain->phi_host,
                             chain->m * chain->n * sizeof(int),
                             cudaMemcpyHostToDevice));
}

/*---------------------------------------------------------------------------*/
__host__ void copy_slip_springed_chain_device_to_host(slip_springed_chain *chain)
{
    cutilSafeCall(cudaMemcpy(chain->r_host, chain->r_device,
                             3 * chain->m * chain->n * sizeof(float),
                             cudaMemcpyDeviceToHost));
    cutilSafeCall(cudaMemcpy(chain->a_host, chain->a_device,
                             3 * chain->m * chain->n * sizeof(float),
                             cudaMemcpyDeviceToHost));
    cutilSafeCall(cudaMemcpy(chain->phi_host, chain->phi_device,
                             chain->m * chain->n * sizeof(int),
                             cudaMemcpyDeviceToHost));
}

/*---------------------------------------------------------------------------*/
