/*---------------------------------------------------------------------------*/

/*!  \file  rng_gpu.cu
     \brief  pseudo random number generator on GPU  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_RNG_GPU_CU
#define INCLUDE_RNG_GPU_CU

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include <cutil.h>
#include <cutil_inline.h>
#include "rng_gpu.h"

/*---------------------------------------------------------------------------*/
#ifndef M_PIf
#define M_PIf 3.14159265358979f
#endif

// #define USE_LCG

/*---------------------------------------------------------------------------*/
__host__ rng_state_gpu *allocate_rng_state_gpu(int n)
{
    rng_state_gpu *rng_gpu;

    rng_gpu = (rng_state_gpu *)malloc(sizeof(rng_state_gpu));

    rng_gpu->n = n;
    rng_gpu->state_host = (unsigned int *)malloc(n * sizeof(unsigned int));
    cutilSafeCall(cudaMalloc((void **)&(rng_gpu->state_device),
                             n * sizeof(unsigned int)));

    return rng_gpu;
}

/*---------------------------------------------------------------------------*/
__host__ void free_rng_state_gpu(rng_state_gpu *rng_gpu)
{
    cutilSafeCall(cudaFree(rng_gpu->state_device));
    free(rng_gpu->state_host);
    free(rng_gpu);
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_rng_state_gpu(rng_state_gpu *rng_gpu,
                                       unsigned int seed)
{
    int i;

    rng_gpu->state_host[0] = seed;
    for(i = 1;i < rng_gpu->n;i++)
    {
        rng_gpu->state_host[i] =
            SRAND_GPU_A * rng_gpu->state_host[i - 1] + SRAND_GPU_C;
    }
}

/*---------------------------------------------------------------------------*/
__host__ void copy_rng_state_gpu_host_to_device(rng_state_gpu *rng_gpu)
{
    cutilSafeCall(cudaMemcpy(rng_gpu->state_device, rng_gpu->state_host,
                             rng_gpu->n * sizeof(unsigned int),
                             cudaMemcpyHostToDevice));
}

/*---------------------------------------------------------------------------*/
__host__ void copy_rng_state_gpu_device_to_host(rng_state_gpu *rng_gpu)
{
    cutilSafeCall(cudaMemcpy(rng_gpu->state_host, rng_gpu->state_device,
                             rng_gpu->n * sizeof(unsigned int),
                             cudaMemcpyDeviceToHost));
}

/*---------------------------------------------------------------------------*/
__device__ int rand_gpu(unsigned int *s)
{
#ifdef USE_LCG
        /*  lineare congruential method  */
    *s = LCG_GPU_A * (*s) + LCG_GPU_C;
#else
        /*  Marsaglia's xor-shift method
            [G. Marsaglia, J. Stat Software 8(14), 1-6 (2003)]  */
    *s ^= ((*s) << XORSHIFT_GPU_C);
    *s ^= ((*s) >> XORSHIFT_GPU_B);
    *s ^= ((*s) << XORSHIFT_GPU_A);
#endif

    return ((*s) >> 1);
}

/*---------------------------------------------------------------------------*/
__device__ float urand_gpu(unsigned int *s)
{
    return (rand_gpu(s) / ((float)(RAND_GPU_MAX + 1)));
}

/*---------------------------------------------------------------------------*/
__device__ void nrand2_gpu(unsigned int *s,float *w1,float *w2)
{
    float u1,u2;
    float r,theta;

    u1 = ((unsigned int)rand_gpu(s) + 1) / ((float)(RAND_GPU_MAX + 1));
    u2 = ((unsigned int)rand_gpu(s) + 1) / ((float)(RAND_GPU_MAX + 1));

    r = sqrtf(-2.0f * logf(u1));
    theta = 2.0f * M_PIf * u2;

    *w1 = r * cosf(theta);
    *w2 = r * sinf(theta);
}

/*---------------------------------------------------------------------------*/
__device__ void load_rand_gpu(const int thread_id,unsigned int *rng_state,
                              unsigned int *s)
{
    *s = rng_state[thread_id];
}

/*---------------------------------------------------------------------------*/
__device__ void save_rand_gpu(const int thread_id,unsigned int *rng_state,
                              unsigned int *s)
{
    rng_state[thread_id] = *s;
}

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
