/*---------------------------------------------------------------------------*/

/*!  \file  simulator_gpu.cu
     \brief  kernels for the simulator on GPU  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
__device__ inline void integrate_langevin_equation(float *r,float *a,
                                                   float *f,float *w)
{
    f[0] += kappa[0][0] * r[0] + kappa[0][1] * r[1] + kappa[0][2] * r[2];
    f[1] += kappa[1][0] * r[0] + kappa[1][1] * r[1] + kappa[1][2] * r[2];
    f[2] += kappa[2][0] * r[0] + kappa[2][1] * r[1] + kappa[2][2] * r[2];
    r[0] += dt * f[0] + sqrtf(2.0f * dt) * w[0];
    r[1] += dt * f[1] + sqrtf(2.0f * dt) * w[1];
    r[2] += dt * f[2] + sqrtf(2.0f * dt) * w[2];
     __syncthreads();

    f[0] = kappa[0][0] * a[0] + kappa[0][1] * a[1] + kappa[0][2] * a[2];
    f[1] = kappa[1][0] * a[0] + kappa[1][1] * a[1] + kappa[1][2] * a[2];
    f[2] = kappa[2][0] * a[0] + kappa[2][1] * a[1] + kappa[2][2] * a[2];
    a[0] += dt * f[0];
    a[1] += dt * f[1];
    a[2] += dt * f[2];
}

/*---------------------------------------------------------------------------*/
__device__ inline void reconstruct_slip_spring(int *phi,float *a,
                                               float *r,float *w,
                                               unsigned int *rng)
{
    float u;

    u = urand_gpu(rng);

    if(*phi == 0)
    {
        if(u < psi_plus)
        {
            (*phi) = 1;

            nrand2_gpu(rng,&(w[0]),&(w[1]));
            calculate_new_slip_spring_position(r,w,a);
        }
    }
    else
    {
        if(u < psi_minus)
        {
            (*phi) = 0;
            a[0] = 0.0f;
            a[1] = 0.0f;
            a[2] = 0.0f;
        }
        else if(u < psi_minus_or_rs)
        {
            nrand2_gpu(rng,&(w[0]),&(w[1]));
            calculate_new_slip_spring_position(r,w,a);
        }
    }
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_exchange_probability(float *r0,float *r1,float *a0,
                                               float *a1,int phi0,int phi1,
                                               float *psi_ex)
{
    int i;
    float delta_f_ex;

    delta_f_ex = 0.0f;
    for(i = 0;i < 3;i++)
    {
        delta_f_ex += (r1[i] - r0[i])
            * ((phi0 - phi1) * (r0[i] + r1[i])
               + 2.0f * (phi1 * a1[i] - phi0 * a0[i]));
    }

    *psi_ex = psi0 * (1.0f - tanhf(0.75f * delta_f_ex / n_s));
}

/*---------------------------------------------------------------------------*/
__device__ inline void exchange_slip_springs(float *a0,float *a1,int *phi0,int *phi1,
                                      float psi_ex,unsigned int *rng)
{
    float a_save[3];
    int phi_save;

    if(urand_gpu(rng) < psi_ex)
    {
        a_save[0] = a1[0];
        a_save[1] = a1[1];
        a_save[2] = a1[2];
        phi_save = *phi1;

        a1[0] = a0[0];
        a1[1] = a0[1];
        a1[2] = a0[2];
        *phi1 = *phi0;

        a0[0] = a_save[0];
        a0[1] = a_save[1];
        a0[2] = a_save[2];
        *phi0 = phi_save;
    }
}

/*---------------------------------------------------------------------------*/
__global__ void perform_simulation_gpu(float *r_device,float *a_device,
                                       int *phi_device,unsigned int *rng_state)
{
    const int thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const int local_id = threadIdx.x;
    int index;
    int i,i_plus,i_minus,k;
    float w[4];
    float f[3];
    float psi_ex;
    unsigned int rng;
    __shared__ float r[N_THREADS][3];
    __shared__ float a[N_THREADS][3];
    __shared__ int phi[N_THREADS];
    float (*r_sub)[3];


    load_rand_gpu(thread_id,rng_state,&rng);

    r_sub = r;
    i = local_id;
    while(i >= n)
    {
        r_sub += n;
        i -= n;
    }
    i_plus = ((i == (n - 1)) ? (n - 1) : (i + 1));
    i_minus = ((i == 0) ? (0) : (i - 1));


    for(index = thread_id;index < m * n;index += N_BLOCKS * N_THREADS)
    {
            /*  load data from global memory  */

        r[local_id][0] = r_device[index * 3];
        r[local_id][1] = r_device[index * 3 + 1];
        r[local_id][2] = r_device[index * 3 + 2];
        a[local_id][0] = a_device[index * 3];
        a[local_id][1] = a_device[index * 3 + 1];
        a[local_id][2] = a_device[index * 3 + 2];
        phi[local_id] = phi_device[index];
        __syncthreads();



            /*  main iteration loop  */

        for(k = 0;k < iterations;k++)
        {
                /*  calculation of force  */
            
            calculate_chain_spring_force(r_sub[i],r_sub[i_minus],
                                         r_sub[i_plus],f);

            f[0] += 3.0f / n_s * phi[local_id]
                * (a[local_id][0] - r[local_id][0]);
            f[1] += 3.0f / n_s * phi[local_id]
                * (a[local_id][1] - r[local_id][1]);
            f[2] += 3.0f / n_s * phi[local_id]
                * (a[local_id][2] - r[local_id][2]);


                /*  w[0:2] : for the integration of the Langevin equation
                    w[3] : for the (re)construction of the slip-spring  */
            nrand2_gpu(&rng,&(w[0]),&(w[1]));
            nrand2_gpu(&rng,&(w[2]),&(w[3]));


                /*  integration of the Langevin equation  */
            integrate_langevin_equation(r[local_id],a[local_id],f,w);


            __syncthreads();

            if(((i & 1) == 0) && (i < (n - 1)))
            {
                calculate_exchange_probability(r[local_id],r[local_id + 1],
                                               a[local_id],a[local_id + 1],
                                               phi[local_id],phi[local_id + 1],
                                               &psi_ex);
                exchange_slip_springs(a[local_id],a[local_id + 1],
                                      &(phi[local_id]),&(phi[local_id + 1]),
                                      psi_ex,&rng);
            }
//             __syncthreads();

            if(((i & 1) == 1) && (i < (n - 1)))
            {
                calculate_exchange_probability(r[local_id],r[local_id + 1],
                                               a[local_id],a[local_id + 1],
                                               phi[local_id],phi[local_id + 1],
                                               &psi_ex);
                exchange_slip_springs(a[local_id],a[local_id + 1],
                                      &(phi[local_id]),&(phi[local_id + 1]),
                                      psi_ex,&rng);
            }
//             __syncthreads();


                /*  reconstruct slip-springs  */
            if((i == 0) || (i == (n - 1)))
            {
                reconstruct_slip_spring(&(phi[local_id]),a[local_id],
                                        r[local_id],w,&rng);
            }
        }


            /*  store data to global memory  */

        r_device[index * 3] = r[local_id][0];
        r_device[index * 3 + 1] = r[local_id][1];
        r_device[index * 3 + 2] = r[local_id][2];
        a_device[index * 3] = a[local_id][0];
        a_device[index * 3 + 1] = a[local_id][1];
        a_device[index * 3 + 2] = a[local_id][2];
        phi_device[index] = phi[local_id];
        __syncthreads();
    }


    save_rand_gpu(thread_id,rng_state,&rng);
}

/*---------------------------------------------------------------------------*/
