/*---------------------------------------------------------------------------*/

/*!  \file  dynamics.c
     \brief  time evolution routines for a slip-springed chain.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2010,2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <math.h>
#include <stdio.h>
#include "dynamics.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
void integrate_langevin_equation(slip_springed_chain *chain,rng_state *rng,
                                 double dt,tensor3 kappa)
{
#ifdef USE_EXPLICIT_EULER
    int i,j,n;
    vector3 *r,*a,*f,*w;
    int *phi;
    vector3 dr,da;
    double n_s_inv;

    n = chain->n;
    n_s_inv = 1.0 / chain->n_s_bar;

    r = chain->r;
    a = chain->a;
    phi = chain->phi;
    f = chain->f;
    w = chain->w;

    for(j = 0;j < chain->m;j++)
    {
        for(i = 0;i < n;i++)
        {
            w[i][0] = nrand(rng);
            w[i][1] = nrand(rng);
            w[i][2] = nrand(rng);
        }

        for(i = 0;i < n;i++)
        {
            f[i][0] = 0;
            f[i][1] = 0;
            f[i][2] = 0;
        }


        f[0][0] = 3 * (r[1][0] - r[0][0]);
        f[0][1] = 3 * (r[1][1] - r[0][1]);
        f[0][2] = 3 * (r[1][2] - r[0][2]);
        for(i = 1;i < (n - 1);i++)
        {
            f[i][0] = 3 * (r[i + 1][0] + r[i - 1][0] - 2 * r[i][0]);
            f[i][1] = 3 * (r[i + 1][1] + r[i - 1][1] - 2 * r[i][1]);
            f[i][2] = 3 * (r[i + 1][2] + r[i - 1][2] - 2 * r[i][2]);
        }
        f[n - 1][0] = 3 * (r[n - 2][0] - r[n - 1][0]);
        f[n - 1][1] = 3 * (r[n - 2][1] - r[n - 1][1]);
        f[n - 1][2] = 3 * (r[n - 2][2] - r[n - 1][2]);


        for(i = 0;i < n;i++)
        {
            if(phi[i] == 1)
            {
                f[i][0] += 3 * n_s_inv *(a[i][0] - r[i][0]);
                f[i][1] += 3 * n_s_inv *(a[i][1] - r[i][1]);
                f[i][2] += 3 * n_s_inv *(a[i][2] - r[i][2]);
            }
        }


        for(i = 0;i < n;i++)
        {
            dr[0] = sqrt(2 * dt) * w[i][0] + dt * f[i][0]
                + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                          + kappa[0][2] * r[i][2]);
            dr[1] = sqrt(2 * dt) * w[i][1] + dt * f[i][1]
                + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                          + kappa[1][2] * r[i][2]);
            dr[2] = sqrt(2 * dt) * w[i][2] + dt * f[i][2]
                + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                          + kappa[2][2] * r[i][2]);

            r[i][0] += dr[0];
            r[i][1] += dr[1];
            r[i][2] += dr[2];


            if(phi[i] == 1)
            {
                da[0] = dt * (kappa[0][0] * a[i][0] + kappa[0][1] * a[i][1]
                              + kappa[0][2] * a[i][2]);
                da[1] = dt * (kappa[1][0] * a[i][0] + kappa[1][1] * a[i][1]
                              + kappa[1][2] * a[i][2]);
                da[2] = dt * (kappa[2][0] * a[i][0] + kappa[2][1] * a[i][1]
                              + kappa[2][2] * a[i][2]);

                a[i][0] += da[0];
                a[i][1] += da[1];
                a[i][2] += da[2];
            }
        }

        r += n;
        a += n;
        phi += n;
    }
#else
    int i,j,n;
    vector3 *r,*a,*f,*w;
    int *phi;
    vector3 dr,da,q;
    double n_s,n_s_inv;

    n = chain->n;
    n_s = chain->n_s_bar;
    n_s_inv = 1.0 / chain->n_s_bar;

    r = chain->r;
    a = chain->a;
    phi = chain->phi;
    f = chain->f;
    w = chain->w;

    for(j = 0;j < chain->m;j++)
    {
        for(i = 0;i < n;i++)
        {
            w[i][0] = nrand(rng);
            w[i][1] = nrand(rng);
            w[i][2] = nrand(rng);
        }

        for(i = 0;i < n;i++)
        {
            f[i][0] = 0;
            f[i][1] = 0;
            f[i][2] = 0;
        }


        f[0][0] = 3 * (r[1][0] - r[0][0]);
        f[0][1] = 3 * (r[1][1] - r[0][1]);
        f[0][2] = 3 * (r[1][2] - r[0][2]);
        for(i = 1;i < (n - 1);i++)
        {
            f[i][0] = 3 * (r[i + 1][0] + r[i - 1][0] - 2 * r[i][0]);
            f[i][1] = 3 * (r[i + 1][1] + r[i - 1][1] - 2 * r[i][1]);
            f[i][2] = 3 * (r[i + 1][2] + r[i - 1][2] - 2 * r[i][2]);
        }
        f[n - 1][0] = 3 * (r[n - 2][0] - r[n - 1][0]);
        f[n - 1][1] = 3 * (r[n - 2][1] - r[n - 1][1]);
        f[n - 1][2] = 3 * (r[n - 2][2] - r[n - 1][2]);


        for(i = 0;i < n;i++)
        {
            if(phi[i] == 0)
            {
                dr[0] = sqrt(2 * dt) * w[i][0] + dt * f[i][0]
                    + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                            + kappa[0][2] * r[i][2]);
                dr[1] = sqrt(2 * dt) * w[i][1] + dt * f[i][1]
                    + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                            + kappa[1][2] * r[i][2]);
                dr[2] = sqrt(2 * dt) * w[i][2] + dt * f[i][2]
                    + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                            + kappa[2][2] * r[i][2]);

                r[i][0] += dr[0];
                r[i][1] += dr[1];
                r[i][2] += dr[2];
            }
            else
            {
                q[0] = r[i][0] - a[i][0];
                q[1] = r[i][1] - a[i][1];
                q[2] = r[i][2] - a[i][2];

                da[0] = dt * (kappa[0][0] * a[i][0] + kappa[0][1] * a[i][1]
                              + kappa[0][2] * a[i][2]);
                da[1] = dt * (kappa[1][0] * a[i][0] + kappa[1][1] * a[i][1]
                              + kappa[1][2] * a[i][2]);
                da[2] = dt * (kappa[2][0] * a[i][0] + kappa[2][1] * a[i][1]
                              + kappa[2][2] * a[i][2]);

                a[i][0] += da[0];
                a[i][1] += da[1];
                a[i][2] += da[2];

                r[i][0] = a[i][0] + exp(- 3 * n_s_inv * dt) * q[0]
                    + sqrt((1 - exp(- 6 * n_s_inv * dt))
                           * (n_s / 3)) * w[i][0]
                    + (1 - exp(-3 * n_s_inv * dt)) * (n_s / 3)
                    * (f[i][0] + kappa[0][0] * q[0] + kappa[0][1] * q[1]
                       + kappa[0][2] * q[2]);
                r[i][1] = a[i][1] + exp(- 3 * n_s_inv * dt) * q[1]
                    + sqrt((1 - exp(- 6 * n_s_inv * dt))
                           * (n_s / 3)) * w[i][1]
                    + (1 - exp(-3 * n_s_inv * dt)) * (n_s / 3)
                    * (f[i][1] + kappa[1][0] * q[0] + kappa[1][1] * q[1]
                       + kappa[1][2] * q[2]);
                r[i][2] = a[i][2] + exp(- 3 * n_s_inv * dt) * q[2]
                    + sqrt((1 - exp(- 6 * n_s_inv * dt))
                           * (n_s / 3)) * w[i][2]
                    + (1 - exp(-3 * n_s_inv * dt)) * (n_s / 3)
                    * (f[i][2] + kappa[2][0] * q[0] + kappa[2][1] * q[1]
                       + kappa[2][2] * q[2]);
            }
        }

        r += n;
        a += n;
        phi += n;
    }
#endif
}

/*---------------------------------------------------------------------------*/
void exchange_slip_springs(slip_springed_chain *chain,rng_state *rng,
                           double dt)
{
    int i,j,k,n;
    vector3 *r,*a;
    int *phi;
    double n_s,n0,psi0;
    double delta_f_ex,psi_ex;
    vector3 a_tmp;
    int phi_tmp;

#define FORCE_EX(i,k) ((phi[(i)] - phi[(i) + 1]) \
                       * (r[(i)][(k)] + r[(i + 1)][(k)]) \
                       + 2.0 * (phi[(i) + 1] * a[(i) + 1][(k)]\
                                - phi[(i)] * a[(i)][(k)]))

    n = chain->n;
    n_s = chain->n_s_bar;
    n0 = chain->n0;

        /*  constant exchange rate model  */
/*     psi0 = 1.0 - exp(- 2.0 * dt * chain->zeta_s_inv); */

        /*  constant net exchange rate model  */
    psi0 = 1.0 - exp(- 2.0 * dt * chain->zeta_s_inv
                     * n0 * (1.0 - exp(-1.0 / n0)));

    r = chain->r;
    a = chain->a;
    phi = chain->phi;


    for(j = 0;j < chain->m;j++)
    {
        for(k = 0;k < 2;k++)
        {
            for(i = k;i < (n - 1);i += 2)
            {
                delta_f_ex = (1.5 / n_s)
                    * ((r[i + 1][0] - r[i][0]) * FORCE_EX(i,0)
                       + (r[i + 1][1] - r[i][1]) * FORCE_EX(i,1)
                       + (r[i + 1][2] - r[i][2]) * FORCE_EX(i,2));
                psi_ex = 0.5 * (1.0 - tanh(0.5 * delta_f_ex)) * psi0;

                if(urand(rng) < psi_ex)
                {
                    a_tmp[0] = a[i][0];
                    a_tmp[1] = a[i][1];
                    a_tmp[2] = a[i][2];
                    phi_tmp = phi[i];

                    a[i][0] = a[i + 1][0];
                    a[i][1] = a[i + 1][1];
                    a[i][2] = a[i + 1][2];
                    phi[i] = phi[i + 1];

                    a[i + 1][0] = a_tmp[0];
                    a[i + 1][1] = a_tmp[1];
                    a[i + 1][2] = a_tmp[2];
                    phi[i + 1] = phi_tmp;
                }
            }
        }

        r += n;
        a += n;
        phi += n;
    }

#undef FORCE_EX
}

/*---------------------------------------------------------------------------*/
void reconstruct_slip_springs(slip_springed_chain *chain,rng_state *rng,
                              double dt)
{
    int j,n;
    vector3 *r,*a;
    int *phi;
    double n0,q0,psi_plus,psi_minus,psi_minus_or_rs,psi0;
    double u;

    n = chain->n;

    r = chain->r;
    a = chain->a;
    phi = chain->phi;

    n0 = chain->n0;
    q0 = sqrt(chain->n_s_bar / 3.0);


        /*   maximum destruction rate model  */
/*     psi_minus = exp(- 1.0 / n0) */
/*         * (1.0 - exp(- dt * chain->zeta_s_inv */
/*                      * exp(2.0 / n0) / (exp(1.0 / n0) - 1.0))); */
/*     psi_plus = (1.0 - exp(- 1.0 / n0)) */
/*         * (1.0 - exp(- dt * chain->zeta_s_inv */
/*                      * exp(2.0 / n0) / (exp(1.0 / n0) - 1.0))); */
/*     psi_minus_or_rs = 1.0 - exp(- (dt * chain->zeta_s_inv) */
/*                                 / (n0 * (1.0 - exp(-1.0 / n0)))); */


        /*  constant destruction rate model  */
/*     psi_minus = exp(- 1.0 / n0) */
/*         * (1.0 - exp(- dt * chain->zeta_s_inv * exp(1.0 / n0))); */
/*     psi_plus = (1.0 - exp(- 1.0 / n0)) */
/*         * (1.0 - exp(- dt * chain->zeta_s_inv * exp(1.0 / n0))); */
/*     psi_minus_or_rs = 1.0 - exp(- (dt * chain->zeta_s_inv)); */


        /*  constant net destruction rate model  */
    psi0 = 1.0 - exp(- dt * chain->zeta_s_inv * n0 * (exp(1.0 / n0) - 1.0));
    psi_minus = exp(- 1.0 / n0) * psi0;
    psi_plus = (1.0 - exp(- 1.0 / n0)) * psi0;
    psi_minus_or_rs = 1.0 - exp(- dt * chain->zeta_s_inv
                                * n0 * (1.0 - exp(-1.0 / n0)));


    for(j = 0;j < chain->m;j++)
    {
        u = urand(rng);
        if(phi[0] == 0)
        {
            if(u < psi_plus)
            {
                phi[0] = 1;
                a[0][0] = r[0][0] + q0 * nrand(rng);
                a[0][1] = r[0][1] + q0 * nrand(rng);
                a[0][2] = r[0][2] + q0 * nrand(rng);
            }
        }
        else
        {
            if(u < psi_minus)
            {
                phi[0] = 0;
                a[0][0] = 0.0;
                a[0][1] = 0.0;
                a[0][2] = 0.0;
            }
            else if(u < psi_minus_or_rs)
            {
                a[0][0] = r[0][0] + q0 * nrand(rng);
                a[0][1] = r[0][1] + q0 * nrand(rng);
                a[0][2] = r[0][2] + q0 * nrand(rng);
            }
        }

        u = urand(rng);
        if(phi[n - 1] == 0)
        {
            if(u < psi_plus)
            {
                phi[n - 1] = 1;
                a[n - 1][0] = r[n - 1][0] + q0 * nrand(rng);
                a[n - 1][1] = r[n - 1][1] + q0 * nrand(rng);
                a[n - 1][2] = r[n - 1][2] + q0 * nrand(rng);
            }
        }
        else
        {
            if(u < psi_minus)
            {
                phi[n - 1] = 0;
                a[n - 1][0] = 0.0;
                a[n - 1][1] = 0.0;
                a[n - 1][2] = 0.0;
            }
            else if(u < psi_minus_or_rs)
            {
                a[n - 1][0] = r[n - 1][0] + q0 * nrand(rng);
                a[n - 1][1] = r[n - 1][1] + q0 * nrand(rng);
                a[n - 1][2] = r[n - 1][2] + q0 * nrand(rng);
            }
        }

        r += n;
        a += n;
        phi += n;
    }
}

/*---------------------------------------------------------------------------*/
