/*---------------------------------------------------------------------------*/

/*!  \file  frisca_cg.c
     \brief  a coarse-grained slip-spring model simulator.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include "frisca_cg.h"
#include "rng.h"
#include "chain.h"
#include "dynamics.h"
#include "store.h"
#include "load.h"

/*---------------------------------------------------------------------------*/
simulator_state *allocate_simulator_state(int m,int n)
{
    simulator_state *state;

    state = malloc(sizeof(simulator_state));
    state->rng = allocate_rng_state();
    state->chain = allocate_slip_springed_chain(m,n);

    return state;
}

/*---------------------------------------------------------------------------*/
void free_simulator_state(simulator_state *state)
{
    free_slip_springed_chain(state->chain);
    free_rng_state(state->rng);
    free(state);
}

/*---------------------------------------------------------------------------*/
void initialize_simulator_state(simulator_state *state,unsigned long seed,
                                double n0,double n_s,double zeta_s)
{
    initialize_rng_state(state->rng,seed);
    initialize_slip_springed_chain(state->chain,state->rng,n0,n_s,zeta_s);
}

/*---------------------------------------------------------------------------*/
void perform_simulation(simulator_state *state,int iterations,double dt,
                        tensor3 kappa)
{
    int i;

    for(i = 0;i < iterations;i++)
    {
        integrate_langevin_equation(state->chain,state->rng,dt,kappa);
        exchange_slip_springs(state->chain,state->rng,dt);
        reconstruct_slip_springs(state->chain,state->rng,dt);
    }
}

/*---------------------------------------------------------------------------*/
void set_centers_of_mass_to_origin(simulator_state *state)
{
    int i,j,n;
    vector3 rc;
    vector3 *r,*a;

    r = state->chain->r;
    a = state->chain->a;
    n = state->chain->n;

    for(j = 0;j < state->chain->m;j++)
    {
        rc[0] = 0.0;
        rc[1] = 0.0;
        rc[2] = 0.0;
        for(i = 0;i < n;i++)
        {
            rc[0] += r[i][0];
            rc[1] += r[i][1];
            rc[2] += r[i][2];
        }
        rc[0] /= n;
        rc[1] /= n;
        rc[2] /= n;

        for(i = 0;i < n;i++)
        {
            r[i][0] -= rc[0];
            r[i][1] -= rc[1];
            r[i][2] -= rc[2];
            a[i][0] -= rc[0];
            a[i][1] -= rc[1];
            a[i][2] -= rc[2];
        }

        r += n;
        a += n;
    }
}

/*---------------------------------------------------------------------------*/
void get_stress_tensor(simulator_state *state,symmetric_tensor3 sigma)
{
    calculate_average_stress_tensor(state->chain,sigma);
}

/*---------------------------------------------------------------------------*/
void store_simulator_state_to_file(char *h5filename,simulator_state *state)
{
    store_state_to_file(h5filename,state->rng,state->chain);
}

/*---------------------------------------------------------------------------*/
simulator_state *load_simulator_state_from_file(char *h5filename)
{
    simulator_state *state;

    state = malloc(sizeof(simulator_state));

    load_state_from_file(h5filename,&(state->rng),&(state->chain));

    return state;
}

/*---------------------------------------------------------------------------*/
