/*---------------------------------------------------------------------------*/

/*!  \file  load.c
     \brief  load the state from a HDF5 file  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "load.h"
#include <stdio.h>
#include <stdlib.h>
#include <hdf5.h>
#include <hdf5_hl.h>

/*---------------------------------------------------------------------------*/
static rng_state *load_rng_from_hdf5(hid_t file);
static slip_springed_chain *load_slip_springed_chain_from_hdf5(hid_t file);
static void read_dataset_unsignedlong(hid_t location,char *path,
                                      unsigned long *dest);

/*---------------------------------------------------------------------------*/
void load_state_from_file(char *h5filename,rng_state **rng,
                          slip_springed_chain **chain)
{
    hid_t file;

    file = H5Fopen(h5filename,H5F_ACC_RDONLY,H5P_DEFAULT);
    if(file < 0)
    {
        fprintf(stderr,"error: cannot open the input HDF5 file `%s'.\n",
                h5filename);
        exit(EXIT_FAILURE);
    }

    *rng = load_rng_from_hdf5(file);
    *chain = load_slip_springed_chain_from_hdf5(file);

    H5Fclose(file);
}

/*---------------------------------------------------------------------------*/
rng_state *load_rng_from_hdf5(hid_t file)
{
    hid_t group1,group2;
    int next_offset;
    rng_state *rng;

    rng = allocate_rng_state();

    group1 = H5Gopen2(file,"/rng_state",H5P_DEFAULT);
    group2 = H5Gopen2(file,"/rng_state/mt_state",H5P_DEFAULT);

    read_dataset_unsignedlong(file,"/rng_state/mt_state/state",rng->mt->state);

    H5LTread_dataset_int(file,"/rng_state/mt_state/n",&(rng->mt->n));
    H5LTread_dataset_int(file,"/rng_state/mt_state/left",&(rng->mt->left));
    H5LTread_dataset_int(file,"/rng_state/mt_state/initf",&(rng->mt->initf));
    H5LTread_dataset_int(file,"/rng_state/mt_state/next",&next_offset);
    rng->mt->next = rng->mt->state + next_offset;

    H5LTread_dataset_int(file,"/rng_state/nrand_has_saved_number",
                         &(rng->nrand_has_saved_number));
    H5LTread_dataset_double(file,"/rng_state/nrand_w",&(rng->nrand_w));

    H5Gclose(group2);
    H5Gclose(group1);

    return rng;
}

/*---------------------------------------------------------------------------*/
slip_springed_chain *load_slip_springed_chain_from_hdf5(hid_t file)
{
    int m,n;
    slip_springed_chain *chain;
    hid_t group;

    group = H5Gopen2(file,"/slip_springed_chain",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/slip_springed_chain/m",&m);
    H5LTread_dataset_int(file,"/slip_springed_chain/n",&n);
    chain = allocate_slip_springed_chain(m,n);

    H5LTread_dataset_double(file,"/slip_springed_chain/n0",&(chain->n0));
    H5LTread_dataset_double(file,"/slip_springed_chain/n_s_bar",
                            &(chain->n_s_bar));
    H5LTread_dataset_double(file,"/slip_springed_chain/zeta_s_inv",
                            &(chain->zeta_s_inv));


    H5LTread_dataset_int(file,"/slip_springed_chain/phi",chain->phi);

    H5LTread_dataset_double(file,"/slip_springed_chain/r",&(chain->r[0][0]));
    H5LTread_dataset_double(file,"/slip_springed_chain/a",&(chain->a[0][0]));


    H5Gclose(group);

    return chain;
}

/*---------------------------------------------------------------------------*/
void read_dataset_unsignedlong(hid_t location,char *path,unsigned long *dest)
{
    hid_t data;

    data = H5Dopen2(location,path,H5P_DEFAULT);

    H5Dread(data,H5T_NATIVE_ULONG,H5S_ALL,H5S_ALL,H5P_DEFAULT,dest);

    H5Dclose(data);
}

/*---------------------------------------------------------------------------*/
