/*---------------------------------------------------------------------------*/

/*!  \file  store.c
     \brief  store the state to a HDF5 file  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "store.h"
#include <hdf5.h>
#include <hdf5_hl.h>

/*---------------------------------------------------------------------------*/
static void store_rng_to_hdf5(hid_t file,rng_state *rng);
static void store_slip_springed_chain_to_hdf5(hid_t file,
                                              slip_springed_chain *chain);
static void make_dataset_unsignedlong(hid_t localtion,char *path,int rank,
                                      hsize_t *dims,unsigned long *src);

/*---------------------------------------------------------------------------*/
void store_state_to_file(char *h5filename,rng_state *rng,
                         slip_springed_chain *chain)
{
    hid_t file;

    file = H5Fcreate(h5filename,H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);

    store_rng_to_hdf5(file,rng);
    store_slip_springed_chain_to_hdf5(file,chain);

    H5Fclose(file);
}

/*---------------------------------------------------------------------------*/
void store_rng_to_hdf5(hid_t file,rng_state *rng)
{
    hid_t group1,group2;
    hsize_t dims[1];
    int next_offset;

    group1 = H5Gcreate2(file,"/rng_state",H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    group2 = H5Gcreate2(file,"/rng_state/mt_state",H5P_DEFAULT,H5P_DEFAULT,
                        H5P_DEFAULT);

    dims[0] = rng->mt->n;
    make_dataset_unsignedlong(file,"/rng_state/mt_state/state",
                              1,dims,rng->mt->state);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/rng_state/mt_state/n",
                         1,dims,&(rng->mt->n));
    H5LTmake_dataset_int(file,"/rng_state/mt_state/left",
                         1,dims,&(rng->mt->left));
    H5LTmake_dataset_int(file,"/rng_state/mt_state/initf",
                         1,dims,&(rng->mt->initf));
    next_offset = rng->mt->next - rng->mt->state;
    H5LTmake_dataset_int(file,"/rng_state/mt_state/next",
                         1,dims,&next_offset);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/rng_state/nrand_has_saved_number",
                         1,dims,&(rng->nrand_has_saved_number));
    H5LTmake_dataset_double(file,"/rng_state/nrand_w",
                            1,dims,&(rng->nrand_w));

    H5Gclose(group2);
    H5Gclose(group1);
}

/*---------------------------------------------------------------------------*/
void store_slip_springed_chain_to_hdf5(hid_t file,slip_springed_chain *chain)
{
    int n,m;
    hid_t group;
    hsize_t dims[3];

    group = H5Gcreate2(file,"/slip_springed_chain",H5P_DEFAULT,H5P_DEFAULT,
                       H5P_DEFAULT);

    m = chain->m;
    n = chain->n;

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/slip_springed_chain/m",1,dims,&m);
    H5LTmake_dataset_int(file,"/slip_springed_chain/n",1,dims,&n);
    H5LTmake_dataset_double(file,"/slip_springed_chain/n0",1,dims,
                            &(chain->n0));
    H5LTmake_dataset_double(file,"/slip_springed_chain/n_s_bar",1,dims,
                            &(chain->n_s_bar));
    H5LTmake_dataset_double(file,"/slip_springed_chain/zeta_s_inv",1,dims,
                            &(chain->zeta_s_inv));

    dims[0] = m;
    dims[1] = n;
    H5LTmake_dataset_int(file,"/slip_springed_chain/phi",2,dims,
                         chain->phi);

    dims[0] = m;
    dims[1] = n;
    dims[2] = 3;
    H5LTmake_dataset_double(file,"/slip_springed_chain/r",3,dims,
                            &(chain->r[0][0]));
    H5LTmake_dataset_double(file,"/slip_springed_chain/a",3,dims,
                            &(chain->a[0][0]));

    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void make_dataset_unsignedlong(hid_t location,char *path,int rank,
                               hsize_t *dims,unsigned long *src)
{
    hid_t data,space;

    space = H5Screate_simple(rank,dims,NULL);
    data = H5Dcreate2(location,path,H5T_NATIVE_ULONG,space,H5P_DEFAULT,
                      H5P_DEFAULT,H5P_DEFAULT);

    H5Dwrite(data,H5T_NATIVE_ULONG,H5S_ALL,H5S_ALL,H5P_DEFAULT,src);

    H5Dclose(data);
    H5Sclose(space);
}

/*---------------------------------------------------------------------------*/
