/*---------------------------------------------------------------------------*/

/*!  \file  chains.cu
     \brief  definitions and subroutines for linear polymer chains.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <cutil.h>
#include <cutil_inline.h>
#include "chains.h"

/*---------------------------------------------------------------------------*/
__host__ polymer_chains *allocate_polymer_chains(int m,int n)
{
    polymer_chains *chains;

    chains = (polymer_chains *)malloc(sizeof(polymer_chains));

    chains->m = m;
    chains->n = n;

    chains->r_host = (float *)malloc(3 * m * n * sizeof(float));
    cutilSafeCall(cudaMalloc((void **)&(chains->r_device),
                             3 * m * n * sizeof(float)));

    return chains;
}

/*---------------------------------------------------------------------------*/
__host__ void free_polymer_chains(polymer_chains *chains)
{
    cutilSafeCall(cudaFree(chains->r_device));
    free(chains->r_host);
    free(chains);
}

/*---------------------------------------------------------------------------*/
void initialize_polymer_chains(polymer_chains *chains,rng_state *rng)
{
    int i,j;
    float *r;

    for(j = 0;j < chains->m;j++)
    {
        r = &(chains->r_host[j * (3 * chains->n)]);

        r[0] = 0.0f;
        r[1] = 0.0f;
        r[2] = 0.0f;

        for(i = 1;i < chains->n;i++)
        {
            r[3 * i] = r[3 * i - 3] + nrand(rng) / sqrtf(3.0f);
            r[3 * i + 1] = r[3 * i - 2] + nrand(rng) / sqrtf(3.0f);
            r[3 * i + 2] = r[3 * i - 1] + nrand(rng) / sqrtf(3.0f);
        }
    }
}

/*---------------------------------------------------------------------------*/
__host__ void copy_polymer_chains_host_to_device(polymer_chains *chains)
{
    cutilSafeCall(cudaMemcpy(chains->r_device, chains->r_host,
                             3 * chains->m * chains->n * sizeof(float),
                             cudaMemcpyHostToDevice));
}

/*---------------------------------------------------------------------------*/
__host__ void copy_polymer_chains_device_to_host(polymer_chains *chains)
{
    cutilSafeCall(cudaMemcpy(chains->r_host, chains->r_device,
                             3 * chains->m * chains->n * sizeof(float),
                             cudaMemcpyDeviceToHost));
}

/*---------------------------------------------------------------------------*/
