/*---------------------------------------------------------------------------*/

/*!  \file  frisca_gpu.cu
     \brief  utility functions to handle the simulator  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <cutil.h>
#include <cutil_inline.h>
#include "frisca_gpu.h"
#include "threads_gpu.h"
#include "store.h"
#include "load.h"

/*---------------------------------------------------------------------------*/
#include "rng_gpu.cu"
#include "kernel_gpu.cu"

/*---------------------------------------------------------------------------*/
#define MAX_DEVICE_COUNT 16

/*---------------------------------------------------------------------------*/
/*!  \brief  calculate a weight value by using the van der Corput sequence.
     \param  i  index for the van der Corput sequence
     \param  b  base for the van der Corput sequence
     \param  w  total value of weights (0 <= return value < \a w)   */
__host__ static int weight_by_van_der_corput_sequence(int i,int b,int w);

/*---------------------------------------------------------------------------*/
__host__ simulator_state *allocate_simulator_state(int m,int n,int z_max)
{
    int s;
    simulator_state *state;

    if((N_THREADS % n) != 0)
    {
        fprintf(stderr,"error: `N_THREADS' is not a multiple of number of beads per chain `n' (not supported).\n");
        exit(EXIT_FAILURE);
    }
    if((m % N_BLOCKS) != 0)
    {
        fprintf(stderr,"error: number of chains `m' is not a multiple of `N_BLOCKS' (not supported).\n");
        exit(EXIT_FAILURE);
    }
    if(z_max != n)
    {
        fprintf(stderr,"error: maximum number of springs `z_max' is not equal to number of beads `n' (not supported).\n");
        exit(EXIT_FAILURE);
    }

    state = (simulator_state *)malloc(sizeof(simulator_state));

    state->m = m;
    state->n = n;
    state->z_max = z_max;

    state->chains = allocate_polymer_chains(m,n);
    state->springs = allocate_slip_springs(m,z_max);
    state->rng = allocate_rng_state();
    state->rng_gpu = allocate_rng_state_gpu(N_BLOCKS * N_THREADS);

    s = ((m * n) + N_BLOCKS - 1) / N_BLOCKS;
    if(s < 1)
        s = 1;
    cutilSafeCall(cudaMalloc((void **)&(state->sigma_subtotal_device[0]),
                             6 * s * sizeof(float)));
    s = ((m * n) + N_BLOCKS - 1) / N_BLOCKS / N_BLOCKS;
    if(s < 1)
        s = 1;
    cutilSafeCall(cudaMalloc((void **)&(state->sigma_subtotal_device[1]),
                             6 * s * sizeof(float)));

    return state;
}

/*---------------------------------------------------------------------------*/
__host__ void free_simulator_state(simulator_state *state)
{
    free_rng_state(state->rng);
    free_polymer_chains(state->chains);
    free_slip_springs(state->springs);
    free_rng_state_gpu(state->rng_gpu);

    cutilSafeCall(cudaFree(state->sigma_subtotal_device[0]));
    cutilSafeCall(cudaFree(state->sigma_subtotal_device[1]));

    free(state);
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_simulator_state(simulator_state *state,
                                         unsigned long seed,
                                         float n_s,float n0)
{
    initialize_rng_state(state->rng,seed);
    initialize_polymer_chains(state->chains,state->rng);
    initialize_slip_springs(state->springs,state->chains,state->rng,n_s,n0);
    initialize_rng_state_gpu(state->rng_gpu,seed);
}

/*---------------------------------------------------------------------------*/
__host__ void set_simulation_parameters(simulator_state *state,float n_s,
                                        float n0,float zeta_s,float dt,
                                        float kappa[3][3],int iterations)
{
    int i,j;

    state->n0 = n0;
    state->n_s = n_s;
    state->dt = dt;
    state->zeta_s_inv = 1.0f / zeta_s;

    for(i = 0;i < 3;i++)
    {
        for(j = 0;j < 3;j++)
        {
            state->kappa[i][j] = kappa[i][j];
        }
    }
    state->iterations = iterations;
}

/*---------------------------------------------------------------------------*/
__host__ void copy_simulator_state_host_to_device(simulator_state *state)
{
    cutilSafeCall(cudaThreadSynchronize());

    copy_polymer_chains_host_to_device(state->chains);
    copy_slip_springs_host_to_device(state->springs);
    copy_rng_state_gpu_host_to_device(state->rng_gpu);

    cutilSafeCall(cudaThreadSynchronize());

    copy_simulator_state_parameters_host_to_device(state);
}

/*---------------------------------------------------------------------------*/
__host__ void copy_simulator_state_parameters_host_to_device(simulator_state *state)
{
    cutilSafeCall(cudaThreadSynchronize());

    cutilSafeCall(cudaMemcpyToSymbol(m,&(state->m),sizeof(int)));
    cutilSafeCall(cudaMemcpyToSymbol(n,&(state->n),sizeof(int)));
    cutilSafeCall(cudaMemcpyToSymbol(z_max,&(state->z_max),sizeof(int)));
    cutilSafeCall(cudaMemcpyToSymbol(n0,&(state->n0),sizeof(float)));
    cutilSafeCall(cudaMemcpyToSymbol(n_s,&(state->n_s),sizeof(float)));
    cutilSafeCall(cudaMemcpyToSymbol(zeta_s_inv,&(state->zeta_s_inv),
                                     sizeof(float)));
    cutilSafeCall(cudaMemcpyToSymbol(dt,&(state->dt),sizeof(float)));
    cutilSafeCall(cudaMemcpyToSymbol(kappa,state->kappa,9 * sizeof(float)));
    cutilSafeCall(cudaMemcpyToSymbol(iterations,&(state->iterations),
                                     sizeof(int)));

    cutilSafeCall(cudaThreadSynchronize());
}

/*---------------------------------------------------------------------------*/
__host__ void copy_simulator_state_device_to_host(simulator_state *state)
{
    cutilSafeCall(cudaThreadSynchronize());

    copy_polymer_chains_device_to_host(state->chains);
    copy_slip_springs_device_to_host(state->springs);
    copy_rng_state_gpu_device_to_host(state->rng_gpu);

    cutilSafeCall(cudaThreadSynchronize());
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_device(int argc,char **argv)
{
    if(cutCheckCmdLineFlag(argc, (const char**)argv, "device"))
    {
        cutilDeviceInit(argc, argv);
    }
    else
    {
        cudaSetDevice(cutGetMaxGflopsDeviceId());
    }
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_device_with_load_balancing(int id)
{
    int i,n,u;
    int device;
    int accumulated_weight[MAX_DEVICE_COUNT];
    cudaDeviceProp device_properties;

    cutilSafeCallNoSync(cudaGetDeviceCount(&n));

    if(n == 0)
    {
        fprintf(stderr,"error: no device supporting CUDA.");
    }
    else if(n == 1)
    {
        cutilSafeCall(cudaSetDevice(0));
    }


    if(n > MAX_DEVICE_COUNT)
        n = MAX_DEVICE_COUNT;

    for(i = 0;i < n;i++)
    {
        cudaGetDeviceProperties(&device_properties,i);

        accumulated_weight[i] =
            ((i == 0) ? (0) : (accumulated_weight[i - 1]))
            + device_properties.multiProcessorCount
              * (device_properties.clockRate / 1000);
    }


    /*  use the van der Corput sequenc (low-discrepancy sequence)
        to select a device with estimated weights  */

    u = weight_by_van_der_corput_sequence(id,n,accumulated_weight[n - 1]);
    device = 0;

    for(i = 0;i < n;i++)
    {
        if(u < accumulated_weight[i])
        {
            device = i;
            break;
        }
    }

    cutilSafeCall(cudaSetDevice(device));
}

/*---------------------------------------------------------------------------*/
__host__ int weight_by_van_der_corput_sequence(int i,int b,int w)
{
    int u,f;

    u = 0;
    f = b;

    while(i > 0)
    {
        u += ((i % b) * w) / f;
        f *= b;

        i /= b;
    }

    return u;
}

/*---------------------------------------------------------------------------*/
__host__ void finalize_device(void)
{
    cudaThreadExit();
}

/*---------------------------------------------------------------------------*/
__host__ void perform_simulation(simulator_state *state)
{
    cutilSafeCall(cudaThreadSynchronize());

    perform_simulation_gpu<<<N_BLOCKS, N_THREADS>>>
        (state->chains->r_device,state->springs->a_device,
         state->springs->s_device,state->rng_gpu->state_device);
    cutilCheckMsg("kernel function execution failed: perform_simulation_gpu_linear()\n");

    cutilSafeCall(cudaThreadSynchronize());
}

/*---------------------------------------------------------------------------*/
__host__ void calculate_stress_tensor(simulator_state *state)
{
    int j;
    int buffer_index[2];

    cutilSafeCall(cudaThreadSynchronize());

    calculate_subtotal_stress_tensor_gpu<<<N_BLOCKS, N_THREADS>>>
        (state->chains->r_device,state->sigma_subtotal_device[0]);
    cutilCheckMsg("kernel function execution failed: calculate_subtotal_stress_tensor_gpu()\n");

    cutilSafeCall(cudaThreadSynchronize());

    buffer_index[0] = 0;
    buffer_index[1] = 1;

    for(j = (state->m * state->n) / N_THREADS;j > 1;j /= N_THREADS)
    {
        reduce_subtotal_stress_tensor_gpu<<<N_BLOCKS, N_THREADS>>>
            (state->sigma_subtotal_device[buffer_index[0]],
             state->sigma_subtotal_device[buffer_index[1]],j);
        cutilCheckMsg("kernel function execution failed: reduce_subtotal_stress_tensor_gpu()\n");

        cutilSafeCall(cudaThreadSynchronize());

        if(buffer_index[0] == 0)
        {
            buffer_index[0] = 1;
            buffer_index[1] = 0;
        }
        else
        {
            buffer_index[0] = 0;
            buffer_index[1] = 1;
        }
    }

    cutilSafeCall(cudaMemcpy(state->sigma,
                             state->sigma_subtotal_device[buffer_index[0]],
                             6 * sizeof(float),cudaMemcpyDeviceToHost));

    state->sigma[0] /= state->m;
    state->sigma[1] /= state->m;
    state->sigma[2] /= state->m;
    state->sigma[3] /= state->m;
    state->sigma[4] /= state->m;
    state->sigma[5] /= state->m;

    cutilSafeCall(cudaThreadSynchronize());
}

/*---------------------------------------------------------------------------*/
__host__ void set_centers_of_mass_to_origin(simulator_state *state)
{
    cutilSafeCall(cudaThreadSynchronize());

    set_centers_of_mass_to_origin_gpu<<<N_BLOCKS, N_THREADS>>>
        (state->chains->r_device,state->springs->a_device);
    cutilCheckMsg("kernel function execution failed: set_centers_of_mass_to_origin_gpu()\n");

    cutilSafeCall(cudaThreadSynchronize());
}

/*---------------------------------------------------------------------------*/
__host__ void store_simulator_state_to_file(char *h5filename,
                                            simulator_state *state)
{
    store_state_to_file(h5filename,state);
}

/*---------------------------------------------------------------------------*/
__host__ simulator_state *load_simulator_state_from_file(char *h5filename)
{
    return load_state_from_file(h5filename);
}

/*---------------------------------------------------------------------------*/
