/*---------------------------------------------------------------------------*/

/*!  \file  springs.cu
     \brief  definitions and subroutines for slip-springs.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <cutil.h>
#include <cutil_inline.h>
#include "springs.h"

/*---------------------------------------------------------------------------*/
#define TOLERANCE_FOR_Z_MAX 1.0e-4

/*---------------------------------------------------------------------------*/
__host__ slip_springs *allocate_slip_springs(int m,int z_max)
{
    slip_springs *springs;

    springs = (slip_springs *)malloc(sizeof(slip_springs));

    springs->m = m;
    springs->z_max = z_max;

    springs->a_host = (float *)malloc(3 * m * z_max * sizeof(float));
    springs->s_host = (int *)malloc(m * z_max * sizeof(int));
    cutilSafeCall(cudaMalloc((void **)&(springs->a_device),
                             3 * m * z_max * sizeof(float)));
    cutilSafeCall(cudaMalloc((void **)&(springs->s_device),
                             m * z_max * sizeof(int)));

    return springs;
}

/*---------------------------------------------------------------------------*/
__host__ void free_slip_springs(slip_springs *springs)
{
    cutilSafeCall(cudaFree(springs->a_device));
    cutilSafeCall(cudaFree(springs->s_device));
    free(springs->a_host);
    free(springs->s_host);
    free(springs);
}

/*---------------------------------------------------------------------------*/
__host__ void initialize_slip_springs(slip_springs *springs,
                                      polymer_chains *chains,
                                      rng_state *rng,float n_s,float n0)
{
    int i,j,z;
    float *a,*r;
    int *s;

    springs->n_s = n_s;
    springs->n0 = n0;

    for(j = 0;j < springs->m;j++)
    {
        a = &(springs->a_host[j * (3 * springs->z_max)]);
        s = &(springs->s_host[j * springs->z_max]);
        r = &(chains->r_host[j * (3 * chains->n)]);

        z = prand(rng,chains->n / n0);

        for(i = 0;i < z;i++)
        {
            s[i] = (int)(chains->n * urand(rng));
            a[3 * i] = r[3 * s[i]]
                + nrand(rng) * sqrtf(n_s / 3.0f);
            a[3 * i + 1] = r[3 * s[i] + 1]
                + nrand(rng) * sqrtf(n_s / 3.0f);
            a[3 * i + 2] = r[3 * s[i] + 2]
                + nrand(rng) * sqrtf(n_s / 3.0f);
        }
        for(;i < springs->z_max;i++)
        {
            s[i] = NO_BEAD_INDEX;
            a[3 * i] = 0.0f;
            a[3 * i + 1] = 0.0f;
            a[3 * i + 2] = 0.0f;
        }
    }
}

/*---------------------------------------------------------------------------*/
__host__ int calculate_estimate_for_z_max(int n,float n0)
{
    return n;
}

/*---------------------------------------------------------------------------*/
__host__ void copy_slip_springs_host_to_device(slip_springs *springs)
{
    cutilSafeCall(cudaMemcpy(springs->a_device, springs->a_host,
                             3 * springs->m * springs->z_max * sizeof(float),
                             cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(springs->s_device, springs->s_host,
                             springs->m * springs->z_max * sizeof(int),
                             cudaMemcpyHostToDevice));
}

/*---------------------------------------------------------------------------*/
__host__ void copy_slip_springs_device_to_host(slip_springs *springs)
{
    cutilSafeCall(cudaMemcpy(springs->a_host, springs->a_device,
                             3 * springs->m * springs->z_max * sizeof(float),
                             cudaMemcpyDeviceToHost));
    cutilSafeCall(cudaMemcpy(springs->s_host, springs->s_device,
                             springs->m * springs->z_max * sizeof(int),
                             cudaMemcpyDeviceToHost));
}

/*---------------------------------------------------------------------------*/
