/*---------------------------------------------------------------------------*/

/*!  \file  store.c
     \brief  store the state to a HDF5 file  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "store.h"
#include <hdf5.h>
#include <hdf5_hl.h>

/*---------------------------------------------------------------------------*/
static void store_rng_to_hdf5(hid_t file,simulator_state *state);
static void store_rng_gpu_to_hdf5(hid_t file,simulator_state *state);
static void store_polymer_chains_to_hdf5(hid_t file,simulator_state *state);
static void store_slip_springs_to_hdf5(hid_t file,simulator_state *state);
static void make_dataset_unsignedlong(hid_t localtion,char *path,int rank,
                                      hsize_t *dims,unsigned long *src);
static void make_dataset_unsignedint(hid_t localtion,char *path,int rank,
                                     hsize_t *dims,unsigned int *src);

/*---------------------------------------------------------------------------*/
void store_state_to_file(char *h5filename,simulator_state *state)
{
    hid_t file;
    hsize_t dims[1];

    file = H5Fcreate(h5filename,H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/m",1,dims,&(state->m));

    store_rng_to_hdf5(file,state);
    store_rng_gpu_to_hdf5(file,state);
    store_polymer_chains_to_hdf5(file,state);
    store_slip_springs_to_hdf5(file,state);

    H5Fclose(file);
}

/*---------------------------------------------------------------------------*/
void store_rng_to_hdf5(hid_t file,simulator_state *state)
{
    rng_state *rng;
    hid_t group1,group2;
    hsize_t dims[1];
    int next_offset;

    rng = state->rng;

    group1 = H5Gcreate2(file,"/rng_state",H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    group2 = H5Gcreate2(file,"/rng_state/mt_state",H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);

    dims[0] = rng->mt->n;
    make_dataset_unsignedlong(file,"/rng_state/mt_state/state",
                              1,dims,rng->mt->state);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/rng_state/mt_state/n",
                         1,dims,&(rng->mt->n));
    H5LTmake_dataset_int(file,"/rng_state/mt_state/left",
                         1,dims,&(rng->mt->left));
    H5LTmake_dataset_int(file,"/rng_state/mt_state/initf",
                         1,dims,&(rng->mt->initf));
    next_offset = rng->mt->next - rng->mt->state;
    H5LTmake_dataset_int(file,"/rng_state/mt_state/next",
                         1,dims,&next_offset);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/rng_state/nrand_has_saved_number",
                         1,dims,&(rng->nrand_has_saved_number));
    H5LTmake_dataset_double(file,"/rng_state/nrand_w",
                            1,dims,&(rng->nrand_w));

    H5Gclose(group2);
    H5Gclose(group1);
}

/*---------------------------------------------------------------------------*/
void store_rng_gpu_to_hdf5(hid_t file,simulator_state *state)
{
    rng_state_gpu *rng_gpu;
    hid_t group;
    hsize_t dims[1];

    rng_gpu = state->rng_gpu;

    group = H5Gcreate2(file,"/rng_gpu_state",H5P_DEFAULT,H5P_DEFAULT,
                        H5P_DEFAULT);

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/rng_gpu_state/n",1,dims,&(rng_gpu->n));

    dims[0] = rng_gpu->n;
    make_dataset_unsignedint(file,"/rng_gpu_state/state",1,dims,
                             rng_gpu->state_host);

    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void store_polymer_chains_to_hdf5(hid_t file,simulator_state *state)
{
    int i,m,n;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    group = H5Gcreate2(file,"/polymer_chain",H5P_DEFAULT,H5P_DEFAULT,
                       H5P_DEFAULT);

    m = state->m;
    n = state->n;

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/polymer_chain/n",1,dims,&n);

    dims[0] = m;
    dims[1] = n;
    dims[2] = 3;
    space = H5Screate_simple(3,dims,NULL);
    data = H5Dcreate2(file,"/polymer_chain/r",H5T_NATIVE_FLOAT,space,
                      H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    dims[0] = n;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = n;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dwrite(data,H5T_NATIVE_FLOAT,slab,space,
                 H5P_DEFAULT,(state->chains->r_host + 3 * n * i));
    }

    H5Sclose(space);
    H5Sclose(slab);
    H5Dclose(data);

    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void store_slip_springs_to_hdf5(hid_t file,simulator_state *state)
{
    int i,m,z_max;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    group = H5Gcreate2(file,"/slip_springs",H5P_DEFAULT,H5P_DEFAULT,
                       H5P_DEFAULT);

    m = state->m;
    z_max = state->z_max;

    dims[0] = 1;
    H5LTmake_dataset_int(file,"/slip_springs/z_max",1,dims,&z_max);
    H5LTmake_dataset_float(file,"/slip_springs/n0",1,dims,
                           &(state->n0));
    H5LTmake_dataset_float(file,"/slip_springs/n_s",1,dims,
                           &(state->n_s));
    H5LTmake_dataset_float(file,"/slip_springs/zeta_s_inv",1,dims,
                           &(state->zeta_s_inv));

    dims[0] = m;
    dims[1] = z_max;
    dims[2] = 3;
    space = H5Screate_simple(3,dims,NULL);
    data = H5Dcreate2(file,"/slip_springs/a",H5T_NATIVE_FLOAT,space,
                      H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    dims[0] = z_max;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = z_max;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dwrite(data,H5T_NATIVE_FLOAT,slab,space,
                 H5P_DEFAULT,(state->springs->a_host + 3 * z_max * i));
    }

    H5Sclose(space);
    H5Sclose(slab);
    H5Dclose(data);


    dims[0] = m;
    dims[1] = z_max;
    space = H5Screate_simple(2,dims,NULL);
    data = H5Dcreate2(file,"/slip_springs/s",H5T_NATIVE_INT,space,
                      H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    dims[0] = z_max;
    slab = H5Screate_simple(1,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    stride[0] = 1;
    stride[1] = 1;
    count[0] = 1;
    count[1] = z_max;
    block[0] = 1;
    block[1] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dwrite(data,H5T_NATIVE_INT,slab,space,
                 H5P_DEFAULT,(state->springs->s_host + z_max * i));
    }

    H5Sclose(space);
    H5Sclose(slab);
    H5Dclose(data);


    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void make_dataset_unsignedlong(hid_t location,char *path,int rank,
                               hsize_t *dims,unsigned long *src)
{
    hid_t data,space;

    space = H5Screate_simple(rank,dims,NULL);
    data = H5Dcreate2(location,path,H5T_NATIVE_ULONG,space,H5P_DEFAULT,
                      H5P_DEFAULT,H5P_DEFAULT);

    H5Dwrite(data,H5T_NATIVE_ULONG,H5S_ALL,H5S_ALL,H5P_DEFAULT,src);

    H5Dclose(data);
    H5Sclose(space);
}

/*---------------------------------------------------------------------------*/
void make_dataset_unsignedint(hid_t location,char *path,int rank,
                              hsize_t *dims,unsigned int *src)
{
    hid_t data,space;

    space = H5Screate_simple(rank,dims,NULL);
    data = H5Dcreate2(location,path,H5T_NATIVE_UINT,space,H5P_DEFAULT,
                      H5P_DEFAULT,H5P_DEFAULT);

    H5Dwrite(data,H5T_NATIVE_UINT,H5S_ALL,H5S_ALL,H5P_DEFAULT,src);

    H5Dclose(data);
    H5Sclose(space);
}

/*---------------------------------------------------------------------------*/
