/*---------------------------------------------------------------------------*/

/*!  \file  frisca.c
     \brief  a single chain slip-spring model simulator.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include "frisca.h"
#include "rng.h"
#include "chain.h"
#include "springs.h"
#include "integrator.h"
#include "jump.h"
#include "strain.h"
#include "store.h"
#include "load.h"

/*---------------------------------------------------------------------------*/
simulator_state *allocate_simulator_state(int m,int n,int z_max)
{
    int j;
    simulator_state *state;

    state = malloc(sizeof(simulator_state));
    state->m = m;
    state->rng = allocate_rng_state();

    state->chain = malloc(sizeof(polymer_chain *) * m);
    state->springs = malloc(sizeof(slip_springs *) * m);
    for(j = 0;j < m;j++)
    {
        state->chain[j] = allocate_polymer_chain(n);
        state->springs[j] = allocate_slip_springs(z_max);
    }

    return state;
}

/*---------------------------------------------------------------------------*/
void free_simulator_state(simulator_state *state)
{
    int j;

    for(j = 0;j < state->m;j++)
    {
        free_polymer_chain(state->chain[j]);
        free_slip_springs(state->springs[j]);
    }
    free(state->chain);
    free(state->springs);
    free_rng_state(state->rng);
    free(state);
}

/*---------------------------------------------------------------------------*/
void initialize_simulator_state(simulator_state *state,unsigned long seed,
                                double n_s,double n0,double zeta_s)
{
    int j;

    initialize_rng_state(state->rng,seed);

    for(j = 0;j < state->m;j++)
    {
        initialize_polymer_chain(state->chain[j],state->rng,0,NULL);
        initialize_slip_springs(state->springs[j],state->chain[j],
                                state->rng,n_s,n0,zeta_s,1.0 / 0.0);
    }
}

/*---------------------------------------------------------------------------*/
void perform_simulation(simulator_state *state,int iterations,double dt,
                        tensor3 kappa)
{
    int i,j;

    for(i = 0;i < iterations;i++)
    {
        for(j = 0;j < state->m;j++)
        {
            integrate_langevin_equation(state->chain[j],state->springs[j],
                                        state->rng,dt,kappa);
            move_slip_springs_by_jumps(state->chain[j],state->springs[j],
                                       state->rng,dt);
            reconstruct_slip_springs(state->chain[j],state->springs[j],
                                     state->rng,dt);
        }
    }
}

/*---------------------------------------------------------------------------*/
void set_centers_of_mass_to_origin(simulator_state *state)
{
    int i,j;
    vector3 rc;
    vector3 *r,*a;

    for(j = 0;j < state->m;j++)
    {
        calculate_center_of_mass(state->chain[j],rc);

        r = state->chain[j]->r;
        a = state->springs[j]->a;

        for(i = 0;i < state->chain[j]->n;i++)
        {
            r[i][0] -= rc[0];
            r[i][1] -= rc[1];
            r[i][2] -= rc[2];
        }

        for(i = 0;i < state->springs[j]->z;i++)
        {
            if(state->springs[j]->s[i] != NO_BEAD_INDEX)
            {
                a[i][0] -= rc[0];
                a[i][1] -= rc[1];
                a[i][2] -= rc[2];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void get_stress_tensor(simulator_state *state,symmetric_tensor3 sigma)
{
    calculate_average_stress_tensor(state->chain,state->m,sigma);
}

/*---------------------------------------------------------------------------*/
void store_simulator_state_to_file(char *h5filename,simulator_state *state)
{
    store_state_to_file_multi_chains(h5filename,state->m,state->rng,
                                     state->chain,state->springs);
}

/*---------------------------------------------------------------------------*/
simulator_state *load_simulator_state_from_file(char *h5filename)
{
    simulator_state *state;

    state = malloc(sizeof(simulator_state));

    load_state_from_file_multi_chains(h5filename,&(state->m),&(state->rng),
                                      &(state->chain),&(state->springs));

    return state;
}

/*---------------------------------------------------------------------------*/
