/*---------------------------------------------------------------------------*/

/*!  \file  integrator_kernel.c
     \brief  kernel for integrator.c.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#define KERNEL2(x,y) x ## _ ## y
#define KERNEL(x,y) KERNEL2(x,y)

/*---------------------------------------------------------------------------*/
static void KERNEL(integrate_langevin_equation,SUFFIX)(
    polymer_chain *chain,slip_springs *springs,rng_state *rng,double dt,
    tensor3 kappa)
{
#ifdef USE_SRK2
    int i;
    vector3 *r,*f,*r_star,*f_star,*w;
    vector3 *a,*a_star;
    vector3 dr,da;
#ifdef BRANCHED_CHAIN
    double *zeta_b_inv;

    zeta_b_inv = chain->zeta_b_inv;
#endif

    r = chain->r;
    f = chain->f;
    r_star = chain->r_star;
    f_star = chain->f_star;
    w = chain->w;

    a = springs->a;
    a_star = springs->a_star;

    clear_force(chain);

        /*  stochastic Runge-Kutta 1st step  */

    generate_random_force(chain,rng);
    calculate_spring_force(chain,springs,1);

#ifdef BRANCHED_CHAIN
    i = 0;
    r_star[i][0] = r[i][0] + sqrt(2 * zeta_b_inv[0] * dt) * w[i][0]
        + dt * zeta_b_inv[0] * f[i][0]
        + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                + kappa[0][2] * r[i][2]);
    r_star[i][1] = r[i][1] + sqrt(2 * zeta_b_inv[0] * dt) * w[i][1]
        + dt * zeta_b_inv[0] * f[i][1]
        + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                + kappa[1][2] * r[i][2]);
    r_star[i][2] = r[i][2] + sqrt(2 * zeta_b_inv[0] * dt) * w[i][2]
        + dt * zeta_b_inv[0] * f[i][2]
        + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                + kappa[2][2] * r[i][2]);

    for(i = 1;i < (chain->n - 1);i++)
#else
    for(i = 0;i < chain->n;i++)
#endif
    {
        r_star[i][0] = r[i][0] + sqrt(2 * dt) * w[i][0] + dt * f[i][0]
            + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                    + kappa[0][2] * r[i][2]);
        r_star[i][1] = r[i][1] + sqrt(2 * dt) * w[i][1] + dt * f[i][1]
            + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                    + kappa[1][2] * r[i][2]);
        r_star[i][2] = r[i][2] + sqrt(2 * dt) * w[i][2] + dt * f[i][2]
            + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                    + kappa[2][2] * r[i][2]);
    }
#ifdef BRANCHED_CHAIN
    i = chain->n - 1;
    r_star[i][0] = r[i][0] + sqrt(2 * zeta_b_inv[1] * dt) * w[i][0]
        + dt * zeta_b_inv[1] * f[i][0]
        + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                + kappa[0][2] * r[i][2]);
    r_star[i][1] = r[i][1] + sqrt(2 * zeta_b_inv[1] * dt) * w[i][1]
        + dt * zeta_b_inv[1] * f[i][1]
        + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                + kappa[1][2] * r[i][2]);
    r_star[i][2] = r[i][2] + sqrt(2 * zeta_b_inv[1] * dt) * w[i][2]
        + dt * zeta_b_inv[1] * f[i][2]
        + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                + kappa[2][2] * r[i][2]);
#endif

    for(i = 0;i < springs->z;i++)
    {
        a_star[i][0] = a[i][0]
            + dt * (kappa[0][0] * a[i][0] + kappa[0][1] * a[i][1]
                    + kappa[0][2] * a[i][2]);
        a_star[i][1] = a[i][1]
            + dt * (kappa[1][0] * a[i][0] + kappa[1][1] * a[i][1]
                    + kappa[1][2] * a[i][2]);
        a_star[i][2] = a[i][2]
            + dt * (kappa[2][0] * a[i][0] + kappa[2][1] * a[i][1]
                    + kappa[2][2] * a[i][2]);
    }


        /*  stochastic Runge-Kutta 2nd step  */

    calculate_spring_force(chain,springs,2);

#ifdef BRANCHED_CHAIN
    i = 0;
    dr[0] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][0]
        + dt * zeta_b_inv[0] * 0.5 * (f[i][0] + f_star[i][0])
        + dt * 0.5 * (kappa[0][0] * (r[i][0] + r_star[i][0])
                      + kappa[0][1] * (r[i][1] + r_star[i][1])
                      + kappa[0][2] * (r[i][2] + r_star[i][2]));
    dr[1] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][1]
        + dt * zeta_b_inv[0] * 0.5 * (f[i][1] + f_star[i][1])
        + dt * 0.5 * (kappa[1][0] * (r[i][0] + r_star[i][0])
                      + kappa[1][1] * (r[i][1] + r_star[i][1])
                      + kappa[1][2] * (r[i][2] + r_star[i][2]));
    dr[2] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][2]
        + dt * zeta_b_inv[0] * 0.5 * (f[i][2] + f_star[i][2])
        + dt * 0.5 * (kappa[2][0] * (r[i][0] + r_star[i][0])
                      + kappa[2][1] * (r[i][1] + r_star[i][1])
                      + kappa[2][2] * (r[i][2] + r_star[i][2]));

    r[i][0] += dr[0];
    r[i][1] += dr[1];
    r[i][2] += dr[2];

    for(i = 1;i < (chain->n - 1);i++)
#else
    for(i = 0;i < chain->n;i++)
#endif
    {
        dr[0] = sqrt(2 * dt) * w[i][0] + dt * 0.5 * (f[i][0] + f_star[i][0])
            + dt * 0.5 * (kappa[0][0] * (r[i][0] + r_star[i][0])
                          + kappa[0][1] * (r[i][1] + r_star[i][1])
                          + kappa[0][2] * (r[i][2] + r_star[i][2]));
        dr[1] = sqrt(2 * dt) * w[i][1] + dt * 0.5 * (f[i][1] + f_star[i][1])
            + dt * 0.5 * (kappa[1][0] * (r[i][0] + r_star[i][0])
                          + kappa[1][1] * (r[i][1] + r_star[i][1])
                          + kappa[1][2] * (r[i][2] + r_star[i][2]));
        dr[2] = sqrt(2 * dt) * w[i][2] + dt * 0.5 * (f[i][2] + f_star[i][2])
            + dt * 0.5 * (kappa[2][0] * (r[i][0] + r_star[i][0])
                          + kappa[2][1] * (r[i][1] + r_star[i][1])
                          + kappa[2][2] * (r[i][2] + r_star[i][2]));

        r[i][0] += dr[0];
        r[i][1] += dr[1];
        r[i][2] += dr[2];
    }
#ifdef BRANCHED_CHAIN
    i = chain->n - 1;
    dr[0] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][0]
        + dt * zeta_b_inv[1] * 0.5 * (f[i][0] + f_star[i][0])
        + dt * 0.5 * (kappa[0][0] * (r[i][0] + r_star[i][0])
                      + kappa[0][1] * (r[i][1] + r_star[i][1])
                      + kappa[0][2] * (r[i][2] + r_star[i][2]));
    dr[1] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][1]
        + dt * zeta_b_inv[1] * 0.5 * (f[i][1] + f_star[i][1])
        + dt * 0.5 * (kappa[1][0] * (r[i][0] + r_star[i][0])
                      + kappa[1][1] * (r[i][1] + r_star[i][1])
                      + kappa[1][2] * (r[i][2] + r_star[i][2]));
    dr[2] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][2]
        + dt * zeta_b_inv[1] * 0.5 * (f[i][2] + f_star[i][2])
        + dt * 0.5 * (kappa[2][0] * (r[i][0] + r_star[i][0])
                      + kappa[2][1] * (r[i][1] + r_star[i][1])
                      + kappa[2][2] * (r[i][2] + r_star[i][2]));

    r[i][0] += dr[0];
    r[i][1] += dr[1];
    r[i][2] += dr[2];
#endif

    for(i = 0;i < springs->z;i++)
    {
        da[0] = dt * 0.5 * (kappa[0][0] * (a[i][0] + a_star[i][0])
                      + kappa[0][1] * (a[i][1] + a_star[i][1])
                      + kappa[0][2] * (a[i][2] + a_star[i][2]));
        da[1] = dt * 0.5 * (kappa[1][0] * (a[i][0] + a_star[i][0])
                      + kappa[1][1] * (a[i][1] + a_star[i][1])
                      + kappa[1][2] * (a[i][2] + a_star[i][2]));
        da[2] = dt * 0.5 * (kappa[2][0] * (a[i][0] + a_star[i][0])
                      + kappa[2][1] * (a[i][1] + a_star[i][1])
                      + kappa[2][2] * (a[i][2] + a_star[i][2]));
        a[i][0] += da[0];
        a[i][1] += da[1];
        a[i][2] += da[2];
    }

#else
    int i;
    vector3 *r,*f,*w;
    vector3 *a;
    vector3 dr,da;
#ifdef BRANCHED_CHAIN
    double *zeta_b_inv;

    zeta_b_inv = chain->zeta_b_inv;
#endif

    r = chain->r;
    f = chain->f;
    w = chain->w;

    a = springs->a;

    clear_force(chain);

    generate_random_force(chain,rng);

    calculate_spring_force(chain,springs,1);

#ifdef BRANCHED_CHAIN
    i = 0;
    dr[0] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][0]
        + zeta_b_inv[0] * dt * f[i][0]
        + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                  + kappa[0][2] * r[i][2]);
    dr[1] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][1]
        + zeta_b_inv[0] * dt * f[i][1]
        + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                  + kappa[1][2] * r[i][2]);
    dr[2] = sqrt(2 * zeta_b_inv[0] * dt) * w[i][2]
        + zeta_b_inv[0] * dt * f[i][2]
        + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                  + kappa[2][2] * r[i][2]);

    r[i][0] += dr[0];
    r[i][1] += dr[1];
    r[i][2] += dr[2];

    for(i = 1;i < (chain->n - 1);i++)
#else
    for(i = 0;i < chain->n;i++)
#endif
    {
        dr[0] = sqrt(2 * dt) * w[i][0] + dt * f[i][0]
            + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                    + kappa[0][2] * r[i][2]);
        dr[1] = sqrt(2 * dt) * w[i][1] + dt * f[i][1]
            + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                    + kappa[1][2] * r[i][2]);
        dr[2] = sqrt(2 * dt) * w[i][2] + dt * f[i][2]
            + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                    + kappa[2][2] * r[i][2]);

        r[i][0] += dr[0];
        r[i][1] += dr[1];
        r[i][2] += dr[2];
    }
#ifdef BRANCHED_CHAIN
    i = chain->n - 1;
    dr[0] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][0]
        + zeta_b_inv[1] * dt * f[i][0]
        + dt * (kappa[0][0] * r[i][0] + kappa[0][1] * r[i][1]
                  + kappa[0][2] * r[i][2]);
    dr[1] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][1]
        + zeta_b_inv[1] * dt * f[i][1]
        + dt * (kappa[1][0] * r[i][0] + kappa[1][1] * r[i][1]
                  + kappa[1][2] * r[i][2]);
    dr[2] = sqrt(2 * zeta_b_inv[1] * dt) * w[i][2]
        + zeta_b_inv[1] * dt * f[i][2]
        + dt * (kappa[2][0] * r[i][0] + kappa[2][1] * r[i][1]
                  + kappa[2][2] * r[i][2]);

    r[i][0] += dr[0];
    r[i][1] += dr[1];
    r[i][2] += dr[2];
#endif

    for(i = 0;i < springs->z;i++)
    {
        da[0] = dt * (kappa[0][0] * a[i][0] + kappa[0][1] * a[i][1]
                         + kappa[0][2] * a[i][2]);
        da[1] = dt * (kappa[1][0] * a[i][0] + kappa[1][1] * a[i][1]
                      + kappa[1][2] * a[i][2]);
        da[2] = dt * (kappa[2][0] * a[i][0] + kappa[2][1] * a[i][1]
                      + kappa[2][2] * a[i][2]);

        a[i][0] += da[0];
        a[i][1] += da[1];
        a[i][2] += da[2];
    }

#endif
}

/*---------------------------------------------------------------------------*/
#undef KERNEL
#undef KERNEL2

/*---------------------------------------------------------------------------*/
