/*---------------------------------------------------------------------------*/

/*!  \file  load.c
     \brief  load the state from a HDF5 file  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "load.h"
#include <stdio.h>
#include <stdlib.h>
#include <hdf5.h>
#include <hdf5_hl.h>

/*---------------------------------------------------------------------------*/
static rng_state *load_rng_from_hdf5(hid_t file);
static polymer_chain **load_polymer_chain_from_hdf5_multi_chains(hid_t file,
                                                                 int m);
static slip_springs **load_slip_springs_from_hdf5_multi_chains(hid_t file,
                                                               int m);
static polymer_chain *load_polymer_chain_from_hdf5_single_chain(hid_t file);
static slip_springs *load_slip_springs_from_hdf5_single_chain(hid_t file);
static void read_dataset_unsignedlong(hid_t location,char *path,
                                      unsigned long *dest);

/*---------------------------------------------------------------------------*/
void load_state_from_file_multi_chains(char *h5filename,int *m,rng_state **rng,
                                       polymer_chain ***chain,
                                       slip_springs ***springs)
{
    hid_t file;

    file = H5Fopen(h5filename,H5F_ACC_RDONLY,H5P_DEFAULT);
    if(file < 0)
    {
        fprintf(stderr,"error: cannot open the input HDF5 file `%s'.\n",
                h5filename);
        exit(EXIT_FAILURE);
    }

    H5LTread_dataset_int(file,"/m",m);

    *rng = load_rng_from_hdf5(file);
    *chain = load_polymer_chain_from_hdf5_multi_chains(file,*m);
    *springs = load_slip_springs_from_hdf5_multi_chains(file,*m);

    H5Fclose(file);
}

/*---------------------------------------------------------------------------*/
void load_state_from_file_single_chain(char *h5filename,rng_state **rng,
                                       polymer_chain **chain,
                                       slip_springs **springs)
{
    hid_t file;

    file = H5Fopen(h5filename,H5F_ACC_RDONLY,H5P_DEFAULT);

    *rng = load_rng_from_hdf5(file);
    *chain = load_polymer_chain_from_hdf5_single_chain(file);
    *springs = load_slip_springs_from_hdf5_single_chain(file);

    H5Fclose(file);
}

/*---------------------------------------------------------------------------*/
rng_state *load_rng_from_hdf5(hid_t file)
{
    hid_t group1,group2;
    int next_offset;
    rng_state *rng;

    rng = allocate_rng_state();

    group1 = H5Gopen2(file,"/rng_state",H5P_DEFAULT);
    group2 = H5Gopen2(file,"/rng_state/mt_state",H5P_DEFAULT);

    read_dataset_unsignedlong(file,"/rng_state/mt_state/state",rng->mt->state);

    H5LTread_dataset_int(file,"/rng_state/mt_state/n",&(rng->mt->n));
    H5LTread_dataset_int(file,"/rng_state/mt_state/left",&(rng->mt->left));
    H5LTread_dataset_int(file,"/rng_state/mt_state/initf",&(rng->mt->initf));
    H5LTread_dataset_int(file,"/rng_state/mt_state/next",&next_offset);
    rng->mt->next = rng->mt->state + next_offset;

    H5LTread_dataset_int(file,"/rng_state/nrand_has_saved_number",
                         &(rng->nrand_has_saved_number));
    H5LTread_dataset_double(file,"/rng_state/nrand_w",&(rng->nrand_w));

    H5Gclose(group2);
    H5Gclose(group1);

    return rng;
}

/*---------------------------------------------------------------------------*/
polymer_chain **load_polymer_chain_from_hdf5_multi_chains(hid_t file,int m)
{
    int i,n,n_b;
    double zeta_b_inv[2];
    polymer_chain **chain;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    chain = malloc(sizeof(polymer_chain *) * m);

    group = H5Gopen2(file,"/polymer_chain",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/polymer_chain/n",&n);
    H5LTread_dataset_int(file,"/polymer_chain/n_b",&n_b);
    H5LTread_dataset_double(file,"/polymer_chain/zeta_b_inv",zeta_b_inv);

    for(i = 0;i < m;i++)
    {
        chain[i] = allocate_polymer_chain(n);
        chain[i]->n_b = n_b;
        chain[i]->zeta_b_inv[0] = zeta_b_inv[0];
        chain[i]->zeta_b_inv[1] = zeta_b_inv[1];
    }


    data = H5Dopen2(file,"/polymer_chain/r",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = n;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = n;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_DOUBLE,slab,space,
                H5P_DEFAULT,chain[i]->r);
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);

    H5Gclose(group);

    return chain;
}

/*---------------------------------------------------------------------------*/
slip_springs **load_slip_springs_from_hdf5_multi_chains(hid_t file,int m)
{
    int i,z_max;
    double n0,n_s,zeta_s_inv,tau_s_inv;
    slip_springs **springs;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    springs = malloc(sizeof(slip_springs *) * m);

    group = H5Gopen2(file,"/slip_springs",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/slip_springs/z_max",&z_max);
    H5LTread_dataset_double(file,"/slip_springs/n0",&n0);
    H5LTread_dataset_double(file,"/slip_springs/n_s",&n_s);
    H5LTread_dataset_double(file,"/slip_springs/zeta_s_inv",&zeta_s_inv);
    H5LTread_dataset_double(file,"/slip_springs/tau_s_inv",&tau_s_inv);

    for(i = 0;i < m;i++)
    {
        springs[i] = allocate_slip_springs(z_max);
        springs[i]->n0 = n0;
        springs[i]->n_s = n_s;
        springs[i]->zeta_s_inv = zeta_s_inv;
        springs[i]->tau_s_inv = tau_s_inv;
    }


    data = H5Dopen2(file,"/slip_springs/a",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = z_max;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = z_max;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_DOUBLE,slab,space,
                H5P_DEFAULT,springs[i]->a);
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);


    data = H5Dopen2(file,"/slip_springs/s",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = z_max;
    slab = H5Screate_simple(1,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    stride[0] = 1;
    stride[1] = 1;
    count[0] = 1;
    count[1] = z_max;
    block[0] = 1;
    block[1] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_INT,slab,space,
                H5P_DEFAULT,springs[i]->s);
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);


    data = H5Dopen2(file,"/slip_springs/z",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = 1;
    slab = H5Screate_simple(1,dims,NULL);

    start[0] = 0;
    stride[0] = 1;
    count[0] = 1;
    block[0] = 1;
    for(i = 0;i < m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_INT,slab,space,
                H5P_DEFAULT,&(springs[i]->z));
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);


    H5Gclose(group);

    return springs;
}

/*---------------------------------------------------------------------------*/
polymer_chain *load_polymer_chain_from_hdf5_single_chain(hid_t file)
{
    int n;
    polymer_chain *chain;
    hid_t group;

    group = H5Gopen2(file,"/polymer_chain",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/polymer_chain/n",&n);
    chain = allocate_polymer_chain(n);

    H5LTread_dataset_int(file,"/polymer_chain/n_b",&(chain->n_b));
    H5LTread_dataset_double(file,"/polymer_chain/zeta_b_inv",
                            chain->zeta_b_inv);

    H5LTread_dataset_double(file,"/polymer_chain/r",(double *)(chain->r));

    H5Gclose(group);

    return chain;
}

/*---------------------------------------------------------------------------*/
slip_springs *load_slip_springs_from_hdf5_single_chain(hid_t file)
{
    int z_max;
    slip_springs *springs;
    hid_t group;

    group = H5Gopen2(file,"/slip_springs",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/slip_springs/z_max",&z_max);
    springs = allocate_slip_springs(z_max);

    H5LTread_dataset_double(file,"/slip_springs/n0",&(springs->n0));
    H5LTread_dataset_double(file,"/slip_springs/n_s",&(springs->n_s));
    H5LTread_dataset_double(file,"/slip_springs/zeta_s_inv",
                            &(springs->zeta_s_inv));
    H5LTread_dataset_double(file,"/slip_springs/tau_s_inv",
                            &(springs->tau_s_inv));
    H5LTread_dataset_int(file,"/slip_springs/z",&(springs->z));

    H5LTread_dataset_double(file,"/slip_springs/a",(double *)(springs->a));
    H5LTread_dataset_int(file,"/slip_springs/s",springs->s);

    H5Gclose(group);

    return springs;
}

/*---------------------------------------------------------------------------*/
void read_dataset_unsignedlong(hid_t location,char *path,unsigned long *dest)
{
    hid_t data;

    data = H5Dopen2(location,path,H5P_DEFAULT);

    H5Dread(data,H5T_NATIVE_ULONG,H5S_ALL,H5S_ALL,H5P_DEFAULT,dest);

    H5Dclose(data);
}

/*---------------------------------------------------------------------------*/
