/*---------------------------------------------------------------------------*/

/*!  \file  path.c
     \brief  utilities to extract primitive path information.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <math.h>
#include "path.h"
#include "chain.h"
#include "springs.h"

/*---------------------------------------------------------------------------*/
#define SQR(x) ((x) * (x))

/*---------------------------------------------------------------------------*/
static void add_to_histogram_double(double value,double *p,int n,double delta);
static void normalize_histogram_double(double *p,int n,double delta);
static void add_to_histogram_int(int value,double *p,int n);
static void normalize_histogram_int(double *p,int n);

/*---------------------------------------------------------------------------*/
double calculate_primitive_path_length(polymer_chain *chain,
                                       slip_springs *springs)
{
    int i,j,i_prev;
    int is_bound;
    double l;

    l = 0.0;
    i_prev = NO_BEAD_INDEX;

    for(i = 0;i < chain->n;i++)
    {
        is_bound = 0;

        for(j = 0;j < springs->z;j++)
        {
            if(springs->s[j] == i)
            {
                is_bound = 1;
                break;
            }
        }

        if(is_bound)
        {
            if(i_prev != NO_BEAD_INDEX)
            {
                l += sqrt(SQR(chain->r[i][0] - chain->r[i_prev][0])
                          + SQR(chain->r[i][1] - chain->r[i_prev][1])
                          + SQR(chain->r[i][2] - chain->r[i_prev][2]));
            }

            i_prev = i;
        }
    }

    return l;
}

/*---------------------------------------------------------------------------*/
double calculate_average_primitive_path_length(polymer_chain **chain,
                                               slip_springs **springs,
                                               int m)
{
    double l_bar;
    int i;

    l_bar = 0;
    for(i = 0;i < m;i++)
    {
        l_bar += calculate_primitive_path_length(chain[i],springs[i]);
    }
    l_bar /= m;

    return l_bar;
}

/*---------------------------------------------------------------------------*/
void calculate_primitive_path_bond_length_distribution(polymer_chain **chain,
                                                       slip_springs **springs,
                                                       int m,double *p_l,
                                                       int nl,double dl)
{
    double l;
    int i,j;
    vector3 r0,r1;

    for(i = 0;i < m;i++)
    {
        r0[0] = chain[i]->r[0][0];
        r0[1] = chain[i]->r[0][1];
        r0[2] = chain[i]->r[0][2];

        for(j = 0;j < springs[i]->z;j++)
        {
            r1[0] = springs[i]->a[j][0];
            r1[1] = springs[i]->a[j][1];
            r1[2] = springs[i]->a[j][2];

            l = sqrt(SQR(r1[0] - r0[0])
                     + SQR(r1[1] - r0[1])
                     + SQR(r1[2] - r0[2]));
            add_to_histogram_double(l,p_l,nl,dl);

            r0[0] = r1[0];
            r0[1] = r1[1];
            r0[2] = r1[2];
        }

        r1[0] = chain[i]->r[(chain[i]->n) - 1][0];
        r1[1] = chain[i]->r[(chain[i]->n) - 1][1];
        r1[2] = chain[i]->r[(chain[i]->n) - 1][2];

        l = sqrt(SQR(r1[0] - r0[0])
                 + SQR(r1[1] - r0[1])
                 + SQR(r1[2] - r0[2]));
        add_to_histogram_double(l,p_l,nl,dl);
    }

    normalize_histogram_double(p_l,nl,dl);
}

/*---------------------------------------------------------------------------*/
void calculate_primitive_path_monomer_number_distribution(polymer_chain **chain,
                                                          slip_springs **springs,
                                                          int m,double *p_n,
                                                          int nn,double dn)
{
    double l,l_total;
    double n;
    int i,j;
    vector3 r0,r1;

    for(i = 0;i < m;i++)
    {
        l_total = 0;

        r0[0] = chain[i]->r[0][0];
        r0[1] = chain[i]->r[0][1];
        r0[2] = chain[i]->r[0][2];

        for(j = 0;j < springs[i]->z;j++)
        {
            r1[0] = springs[i]->a[j][0];
            r1[1] = springs[i]->a[j][1];
            r1[2] = springs[i]->a[j][2];

            l_total += sqrt(SQR(r1[0] - r0[0])
                            + SQR(r1[1] - r0[1])
                            + SQR(r1[2] - r0[2]));

            r0[0] = r1[0];
            r0[1] = r1[1];
            r0[2] = r1[2];
        }

        r1[0] = chain[i]->r[(chain[i]->n) - 1][0];
        r1[1] = chain[i]->r[(chain[i]->n) - 1][1];
        r1[2] = chain[i]->r[(chain[i]->n) - 1][2];

        l_total += sqrt(SQR(r1[0] - r0[0])
                        + SQR(r1[1] - r0[1])
                        + SQR(r1[2] - r0[2]));


        r0[0] = chain[i]->r[0][0];
        r0[1] = chain[i]->r[0][1];
        r0[2] = chain[i]->r[0][2];

        for(j = 0;j < springs[i]->z;j++)
        {
            r1[0] = springs[i]->a[j][0];
            r1[1] = springs[i]->a[j][1];
            r1[2] = springs[i]->a[j][2];

            l = sqrt(SQR(r1[0] - r0[0])
                     + SQR(r1[1] - r0[1])
                     + SQR(r1[2] - r0[2]));
            n = (chain[i]->n - 1) * l / l_total;
            add_to_histogram_double(n,p_n,nn,dn);

            r0[0] = r1[0];
            r0[1] = r1[1];
            r0[2] = r1[2];
        }

        r1[0] = chain[i]->r[(chain[i]->n) - 1][0];
        r1[1] = chain[i]->r[(chain[i]->n) - 1][1];
        r1[2] = chain[i]->r[(chain[i]->n) - 1][2];

        l = sqrt(SQR(r1[0] - r0[0])
                 + SQR(r1[1] - r0[1])
                 + SQR(r1[2] - r0[2]));
        n = (chain[i]->n - 1) * l / l_total;
        add_to_histogram_double(n,p_n,nn,dn);
    }

    normalize_histogram_double(p_n,nn,dn);
}

/*---------------------------------------------------------------------------*/
void calculate_primitive_path_entanglement_distribution(polymer_chain **chain,
                                                        slip_springs **springs,
                                                        int m,double *p_z,
                                                        int nz)
{
    int i;

    for(i = 0;i < m;i++)
    {
        add_to_histogram_int(springs[i]->z,p_z,nz);
    }

    normalize_histogram_int(p_z,nz);
}

/*---------------------------------------------------------------------------*/
void add_to_histogram_double(double value,double *p,int n,double delta)
{
    int index;

    index = (int)(value / delta);

    if(index < n)
    {
        p[index]++;
    }
}

/*---------------------------------------------------------------------------*/
void normalize_histogram_double(double *p,int n,double delta)
{
    int i;
    double c;

    c = 0;
    for(i = 0;i < n;i++)
    {
        c += p[i];
    }

    for(i = 0;i < n;i++)
    {
        p[i] /= c * delta;
    }
}

/*---------------------------------------------------------------------------*/
void add_to_histogram_int(int value,double *p,int n)
{
    if(value < n)
    {
        p[value]++;
    }
}

/*---------------------------------------------------------------------------*/
void normalize_histogram_int(double *p,int n)
{
    int i;
    double c;

    c = 0;
    for(i = 0;i < n;i++)
    {
        c += p[i];
    }

    for(i = 0;i < n;i++)
    {
        p[i] /= c;
    }
}

/*---------------------------------------------------------------------------*/

