/*---------------------------------------------------------------------------*/

/*!  \file  springs.c
     \brief  definitions and subroutines for slip-springs.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include "springs.h"
#include "mt.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
#define TOLERANCE_FOR_Z_MAX 1.0e-4
#define MAX(x,y) (((x) > (y)) ? (x) : (y))

/*---------------------------------------------------------------------------*/
slip_springs *allocate_slip_springs(int z_max)
{
    slip_springs *springs;

    springs = malloc(sizeof(slip_springs));

    springs->z_max = z_max;

    springs->a = malloc(sizeof(vector3) * z_max);
    springs->a_star = malloc(sizeof(vector3) * z_max);
    springs->s = malloc(sizeof(int) * z_max);

    return springs;
}

/*---------------------------------------------------------------------------*/
void free_slip_springs(slip_springs *springs)
{
    free(springs->a);
    free(springs->a_star);
    free(springs->s);

    free(springs);
}

/*---------------------------------------------------------------------------*/
void initialize_slip_springs(slip_springs *springs,polymer_chain *chain,
                             rng_state *rng,double n_s,double n0,
                             double zeta_s,double tau_s)
{
    int i;

    springs->n_s = n_s;
    springs->n0 = n0;
    springs->zeta_s_inv = 1.0 / zeta_s;
    if(tau_s > 0.0)
        springs->tau_s_inv = 1.0 / tau_s;
    else
        springs->tau_s_inv = 1.0 / 0.0;

    springs->z = prand(rng,chain->n / n0);
    if(springs->z > springs->z_max)
        springs->z = springs->z_max;

    for(i = 0;i < springs->z;i++)
    {
        springs->s[i] = (int)(chain->n * urand(rng));
        springs->a[i][0] = chain->r[springs->s[i]][0]
            + nrand(rng) * sqrt(n_s / 3.0);
        springs->a[i][1] = chain->r[springs->s[i]][1]
            + nrand(rng) * sqrt(n_s / 3.0);
        springs->a[i][2] = chain->r[springs->s[i]][2]
            + nrand(rng) * sqrt(n_s / 3.0);
    }
    for(;i < springs->z_max;i++)
    {
        springs->s[i] = NO_BEAD_INDEX;
        springs->a[i][0] = 0.0;
        springs->a[i][1] = 0.0;
        springs->a[i][2] = 0.0;
    }
}

/*---------------------------------------------------------------------------*/
int calculate_estimate_for_z_max(int n,double n0)
{
    double z_bar;
    int z_estimate;
    int i;
    double p;

    z_bar = (double)n / n0;
    z_estimate = MAX(n, z_bar * 2);

    while(1)
    {
        p = exp(-z_bar);
        for(i = 1;i < z_estimate;i++)
        {
            p *= z_bar / (double)i;
        }

        if(p <= TOLERANCE_FOR_Z_MAX)
            break;

        z_estimate++;
    }

    return z_estimate;
}

/*---------------------------------------------------------------------------*/
void calculate_virtual_stress_tensor(polymer_chain *chain,
                                     slip_springs *springs,
                                     symmetric_tensor3 sigma)
{
    int i,j;
    vector3 *r,*a;
    double n_s;

    r = chain->r;
    a = springs->a;
    n_s = springs->n_s;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;
    for(j = 0;j < springs->z;j++)
    {
        i = springs->s[j];

        sigma[0] += (3.0 / n_s) * (r[i][0] - a[j][0]) * (r[i][0] - a[j][0]);
        sigma[1] += (3.0 / n_s) * (r[i][1] - a[j][1]) * (r[i][1] - a[j][1]);
        sigma[2] += (3.0 / n_s) * (r[i][2] - a[j][2]) * (r[i][2] - a[j][2]);
        sigma[3] += (3.0 / n_s) * (r[i][0] - a[j][0]) * (r[i][1] - a[j][1]);
        sigma[4] += (3.0 / n_s) * (r[i][0] - a[j][0]) * (r[i][2] - a[j][2]);
        sigma[5] += (3.0 / n_s) * (r[i][1] - a[j][1]) * (r[i][2] - a[j][2]);
    }
}

/*---------------------------------------------------------------------------*/
void calculate_average_virtual_stress_tensor(polymer_chain **chain,
                                             slip_springs **springs,
                                             int m,
                                             symmetric_tensor3 sigma)
{
    int j;
    symmetric_tensor3 sigma_1;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;
    for(j = 0;j < m;j++)
    {
        calculate_virtual_stress_tensor(chain[j],springs[j],sigma_1);
        
        sigma[0] += sigma_1[0];
        sigma[1] += sigma_1[1];
        sigma[2] += sigma_1[2];
        sigma[3] += sigma_1[3];
        sigma[4] += sigma_1[4];
        sigma[5] += sigma_1[5];
    }
    sigma[0] /= m;
    sigma[1] /= m;
    sigma[2] /= m;
    sigma[3] /= m;
    sigma[4] /= m;
    sigma[5] /= m;
}

/*---------------------------------------------------------------------------*/
double calculate_average_number_of_slip_springs(slip_springs **springs,int m)
{
    int i;
    double z_bar;

    z_bar = 0;
    for(i = 0;i < m;i++)
    {
        z_bar += springs[i]->z;
    }
    z_bar /= m;

    return z_bar;
}

/*---------------------------------------------------------------------------*/
void sort_slip_springs_by_bead_index(slip_springs *springs)
{
    int i,j,k;
    int s_min;
    vector3 a_min;

    for(i = 0;i < springs->z;i++)
    {
        s_min = springs->s[i];
        k = i;
        for(j = i + 1;j < springs->z;j++)
        {
            if(springs->s[j] < s_min)
            {
                s_min = springs->s[j];
                k = j;
            }
        }

        a_min[0] = springs->a[k][0];
        a_min[1] = springs->a[k][1];
        a_min[2] = springs->a[k][2];

        springs->s[k] = springs->s[i];
        springs->a[k][0] = springs->a[i][0];
        springs->a[k][1] = springs->a[i][1];
        springs->a[k][2] = springs->a[i][2];

        springs->s[i] = s_min;
        springs->a[i][0] = a_min[0];
        springs->a[i][1] = a_min[1];
        springs->a[i][2] = a_min[2];
    }
}

/*---------------------------------------------------------------------------*/
